/* Standard-Definitions */
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "set.h"


/* **********************************************
   Mathe-Funktionen fuer komplexe Zahlen */

/* Load negator bitmask into xmm7 */
   
#ifdef SSE
void init_sse()
 {
  COMPLEX tmp1, tmp2;

  tmp1.re = 1;
  tmp2.re = -tmp1.re;
  tmp1.im = 0;
  tmp2.im = 0;

__asm__ __volatile__ (
	"movapd %0, %%xmm0\n\t"
        "movapd %1, %%xmm7\n\t"
	"xorpd %%xmm0, %%xmm7"
	: "=m" (tmp1.re), "=m" (tmp2.re)
	: "m" (tmp1.re), "m" (tmp2.re) );
 }
#endif

/* Zuerst die vier Grundrechenarten */

#ifndef INLINE 

/* Zwei komplexe Zahlen zueinander addieren */

void add(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest)
 {
   dest->re=source1->re + source2->re;
   dest->im=source1->im + source2->im;
 }

/* Zwei komplexe Zahlen voneinander abziehen */

void sub(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest)
 {
   dest->re=source1->re - source2->re;
   dest->im=source1->im - source2->im;
 }

/* Zwei komplexe Zahlen miteinander multiplizieren */

void multiply(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest)
 {
   dest->re=source1->re * source2->re - source1->im * source2->im;
   dest->im=source1->im * source2->re + source1->re * source2->im;
 }

/* Multiply two complex numbers and add them to destination */

void multiply_add(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest)
 {
   dest->re += source1->re * source2->re - source1->im * source2->im;
   dest->im += source1->im * source2->re + source1->re * source2->im;
 }

/* Multiply two complex numbers and subtract them from destination */

void multiply_sub(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest)
 {
   dest->re -= source1->re * source2->re - source1->im * source2->im;
   dest->im -= source1->im * source2->re + source1->re * source2->im;
 }

#endif

/* Zwei komplexe Zahlen durcheinander dividieren */

void divide(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest)
 {
   REAL div;

   div = source2->re * source2->re + source2->im * source2->im;
   if(! div)
    {
      display_error("Division by zero");
      return;
    }
   dest->re= (source1->re * source2->re + source1->im * source2->im) / div;
   dest->im= (source1->im * source2->re - source1->re * source2->im) / div;
 }

/* Nun ein paar Funktionen fuer komplexe Zahlen */

/* Das Betrags-Quadrat einer komplexen Zahl */

REAL complex_abs_sq(COMPLEX *num)
 {
   REAL value;

   value=num->re * num->re + num->im * num->im;
   return(value);
 }

/* Der Betrag einer komplexen Zahl */

REAL complex_abs(COMPLEX *num)
 {
   REAL value;

   value=num->re * num->re + num->im * num->im;
   return(sqrt(value));
 }

/* Der Realteil einer komplexen Zahl */

REAL complex_re(COMPLEX *num)
 {
   return(num->re);
 }

/* Der Imaginaerteil einer komplexen Zahl */

REAL complex_im(COMPLEX *num)
 {
   return(num->im);
 }

#ifndef INLINE

/* Eine komplexe Zahl konjugieren */

void complex_conjugate(COMPLEX *source, COMPLEX *dest)
 {
   dest->re = source->re;
   dest->im = -source->im;
 }

#endif

/* Das Argument einer komplexen Zahl */

REAL complex_arg(COMPLEX *num)
 {
   REAL value;

   switch(CUT)
    {
      case 1:				/* Schnitt auf der positiven imaginaeren Achse */
       if(num->re == 0.0)		/* Realteil Null behandeln */
        {
         if(num->im < 0.0)
            return(-PI/(2.0));
         return(PI/(2.0));
        }
       value=atan2(num->im, num->re);
       if(value > PI/(2.0))
          value -= 2.0*PI;
       return(value);
      default:
       display_error("Setting of Cut not in the range 1...4");
      case 2:				/* Schnitt auf der negativen reellen Achse */
       if(num->im == 0.0)		/* Imaginaerteil Null behandeln */
        {
         if(num->re < 0.0)
            return(PI);
         return(0.0);
        }
       return(atan2(num->im, num->re));
      case 3:				/* Schnitt auf der negativen imaginaeren Achse */
       if(num->re == 0.0)		/* Realteil Null behandeln */
        {
         if(num->im < 0.0)
            return(-PI/(2.0));
         return(PI/(2.0));
        }
       value=atan2(num->im, num->re);
       if(value < -PI/(2.0))
          value += 2.0*PI;
       return(value);
      case 4:				/* Schnitt auf der positiven reellen Achse */
       if(num->im == 0.0)		/* Imaginaerteil Null behandeln */
        {
         if(num->re < 0.0)
            return(PI);
         return(0.0);
        }
       value=atan2(num->im, num->re);
       if(value < 0.0)
          value += 2.0*PI;
       return(value);
    }
 }

/* Die komplexe Wurzel */

void complex_sqrt(COMPLEX *source, COMPLEX *dest)
 {
   REAL arg, abs;

   if(CUT == 2)				/* Schnitt auf der negativen reellen Achse ? */
    {
     if(source->im == 0.0)		/* Imaginaerteil Null behandeln */
      {
        if(source->re < 0.0)
          {
            dest->re = 0;
            dest->im = sqrt(-source->re);
            return;
          }
        dest->re = sqrt(source->re);
        dest->im = 0;
        return;
      }
    }

   abs=sqrt(complex_abs(source));
   arg=complex_arg(source)/2.0;
   dest->re = abs * cos(arg);
   dest->im = abs * sin(arg);
 }

/* Die Exponential-Funktion */

void complex_exp(COMPLEX *source, COMPLEX *dest)
 {
   REAL ex;

   ex=exp(source->re);
   dest->re=ex*cos(source->im);
   dest->im=ex*sin(source->im);
 }

/* Der komplexe Logarithmus */

void complex_ln(COMPLEX *source, COMPLEX *dest)
 {
   dest->re = log(complex_abs(source));
   dest->im = complex_arg(source);;
 }

/* Komplexe Potenzen */

void complex_pow(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest)
 {
   COMPLEX temp, temp1;

   if((source1->im == 0.0) && (source1->re == 0.0))	/* Sonderbehandlung fuer 0 */
    {
      dest->re=0.0;
      dest->im=0.0;
      return;
    }
   complex_ln(source1, &temp);
   multiply(&temp, source2, &temp1);
   complex_exp(&temp1, dest);
 }

/* Der Sinus */

void complex_sin(COMPLEX *source, COMPLEX *dest)
 {
   REAL ex, exm;

   ex=exp(source->im);
   exm=1.0/ex;
   dest->re=(ex+exm)/2.0*sin(source->re);
   dest->im=(ex-exm)/2.0*cos(source->re);
 }

/* Der Cosinus */

void complex_cos(COMPLEX *source, COMPLEX *dest)
 {
   REAL ex, exm;

   ex=exp(source->im);
   exm=1.0/ex;
   dest->re=(ex+exm)/2.0*cos(source->re);
   dest->im=(exm-ex)/2.0*sin(source->re);
 }

/* Der komplexe Tangens */

void complex_tan(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX sinval, cosval;

   complex_sin(source, &sinval);
   complex_cos(source, &cosval);
   divide(&sinval, &cosval, dest);
 }

/* Der Cotangens */

void complex_cotan(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX sinval, cosval;

   complex_sin(source, &sinval);
   complex_cos(source, &cosval);
   divide(&cosval, &sinval, dest);
 }

/* Der Sinus hyperbolicus */

void complex_sinh(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX pos_exp, neg_exp;
   COMPLEX temp;

   complex_exp(source, &pos_exp);
   temp.re = -source->re;
   temp.im = -source->im;
   complex_exp(&temp, &neg_exp);
   sub(&pos_exp, &neg_exp, &temp);
   dest->re = temp.re/2.0;
   dest->im = temp.im/2.0;
 }

/* Der Cosinus hyperbolicus */

void complex_cosh(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX pos_exp, neg_exp;
   COMPLEX temp;

   complex_exp(source, &pos_exp);
   temp.re = -source->re;
   temp.im = -source->im;
   complex_exp(&temp, &neg_exp);
   add(&pos_exp, &neg_exp, &temp);
   dest->re = temp.re/2.0;
   dest->im = temp.im/2.0;
 }


/* Der komplexe Tangens hyperbolicus */

void complex_tanh(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX sinval, cosval;

   complex_sinh(source, &sinval);
   complex_cosh(source, &cosval);
   divide(&sinval, &cosval, dest);
 }

/* Der Cotangens hyperbolicus */

void complex_cotanh(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX sinval, cosval;

   complex_sinh(source, &sinval);
   complex_cosh(source, &cosval);
   divide(&cosval, &sinval, dest);
 }

/* Der asinh */

void complex_asinh(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX temp, temp1;

   multiply(source, source, &temp);
   temp.re += 1.0;
   complex_sqrt(&temp, &temp1);
   add(&temp1, source, &temp);
   complex_ln(&temp, dest);
 }

/* Der acosh */

void complex_acosh(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX temp, temp1;

   multiply(source, source, &temp);
   temp.re -= 1.0;
   complex_sqrt(&temp, &temp1);
   add(&temp1, source, &temp);
   complex_ln(&temp, dest);
 }

/* Der atanh */

void complex_atanh(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX temp, temp1, temp2;

   temp2.re=1.0;
   temp2.im=0.0;
   add(&temp2, source, &temp);
   sub(&temp2, source, &temp1);
   divide(&temp, &temp1, &temp2);
   complex_sqrt(&temp2, &temp);
   complex_ln(&temp, dest);
 }

/* Der acotanh */

void complex_acotanh(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX temp, temp1, temp2;

   temp2.re=1.0;
   temp2.im=0.0;
   add(&temp2, source, &temp);
   sub(source, &temp2, &temp1);
   divide(&temp, &temp1, &temp2);
   complex_sqrt(&temp2, &temp);
   complex_ln(&temp, dest);
 }

/* Der asin */

void complex_asin(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX temp, temp1;

   temp.re = -source->im;
   temp.im = source->re;
   complex_asinh(&temp, &temp1);
   dest->re = temp1.im;
   dest->im = -temp1.re;
 }

/* Der acos */

void complex_acos(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX temp1;

   complex_acosh(source, &temp1);
   dest->re = temp1.im;
   dest->im = -temp1.re;
 }

/* Der atan */

void complex_atan(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX temp, temp1;

   temp.re = -source->im;
   temp.im = source->re;
   complex_atanh(&temp, &temp1);
   dest->re = temp1.im;
   dest->im = -temp1.re;
 }

/* Der acotan */

void complex_acotan(COMPLEX *source, COMPLEX *dest)
 {
   COMPLEX temp, temp1;

   temp.re = source->im;
   temp.im = -source->re;
   complex_acotanh(&temp, &temp1);
   dest->re = temp1.im;
   dest->im = -temp1.re;
 }

/* Ene komplexe Zahl in einen String verwandeln */

void complex_to_str(COMPLEX *num, UBYTE *str)
 {
   register int i;
   UBYTE hstring[100];

   if(num->im == 0.0)
     sprintf(hstring, "%17.17g", num->re);
   else
    {
      if(num->re == 0.0)
       {
         if(num->im < 0.0)
           sprintf(hstring, "-i*%17.17g", -num->im);
         else
           sprintf(hstring, "i*%17.17g", num->im);
       }
      else
       {
         if(num->im < 0.0)
           sprintf(hstring, "%17.17g-i*%17.17g", num->re, -num->im);
         else
           sprintf(hstring, "%17.17g+i*%17.17g", num->re, num->im);
       }
    }

   for(i=0; hstring[i]; i++)
     if(hstring[i] > ' ')
          *str++ = hstring[i];

   *str=0;
 }

/* Eine komplexe Zahl ausgeben */

void show_complex(COMPLEX *num)
 {
   UBYTE temp_str[100];

   complex_to_str(num, temp_str);
   printf("%s\n", temp_str);
 }

/* Eine komplexe Zahl in eine Datei schreiben */

void output_complex(COMPLEX *num, FILE *fp)
 {
   UBYTE temp_str[100];

   complex_to_str(num, temp_str);
   fprintf(fp, "%s", temp_str);
 }
