/* **********************************************
   November 10-16, 2004:
     some shortcuts to save temporary entries
     (and thus CPU time) added
   ********************************************** */

/* Aligned memory allocation */

#ifdef SSE
#define malign16(a) memalign(16, a) 
#else
#define malign16(a) malloc(a)
#endif

/* Definitions for complex numbers */

/* Type of precision we want to work with */

typedef double REAL;

/* A complex number */

#ifdef SSE
struct _complex { REAL re; REAL im; } __attribute__ ((aligned (16)));
#else
struct _complex { REAL re; REAL im; };
#endif

typedef struct _complex COMPLEX;

/* The constant pi */

#define PI 3.14159265358979323846246

/* The Euler constant */

#define E_CONST 2.718281828459045

/* External functions on complex variables */

/* Basic arithmetics */

#define multiply_add_re(source1, source2, dest) {\
   dest += ((source1)->re * (source2)->re - (source1)->im * (source2)->im); }
#define multiply_add_im(source1, source2, dest) {\
   dest += ((source1)->im * (source2)->re + (source1)->re * (source2)->im); }
#define multiply_sub_re(source1, source2, dest) {\
   dest -= ((source1)->re * (source2)->re - (source1)->im * (source2)->im); }
#define multiply_sub_im(source1, source2, dest) {\
   dest -= ((source1)->im * (source2)->re + (source1)->re * (source2)->im); }
#define multiply_c_add_re(source1, source2, dest) {\
   dest += ((source1)->re * (source2)->re + (source1)->im * (source2)->im); }
#define multiply_c_add_im(source1, source2, dest) {\
   dest += (-(source1)->im * (source2)->re + (source1)->re * (source2)->im); }
/* dest += source1^* * source2 */
#define multiply_c_add(source1, source2, dest) {\
   multiply_c_add_re(source1, source2, (dest)->re); \
   multiply_c_add_im(source1, source2, (dest)->im); }
/* dest += (source1 * source2)^* */
#define multiply_cc_add(source1, source2, dest) {\
   multiply_add_re(source1, source2, (dest)->re); \
   multiply_sub_im(source1, source2, (dest)->im); }

#ifdef INLINE
#ifdef SSE
/* SSE: xmm7 is reserved for negator bitmask */
extern void init_sse();
#define add(source1, source2, dest)	\
__asm__ __volatile__ (			\
	"movapd %0, %%xmm0\n\t"		\
        "movapd %1, %%xmm1\n\t"		\
	"addpd %%xmm0, %%xmm1\n\t"	\
	"movapd %%xmm1, %2"		\
	: "=m" ((source1)->re), "=m" ((source2)->re),  "=m" ((dest)->re)	\
	: "m" ((source1)->re), "m" ((source2)->re) )
#define sub(source1, source2, dest)	\
__asm__ __volatile__ (			\
	"movapd %0, %%xmm0\n\t"		\
        "movapd %1, %%xmm1\n\t"		\
	"subpd %%xmm1, %%xmm0\n\t"	\
	"movapd %%xmm0, %2"		\
	: "=m" ((source1)->re), "=m" ((source2)->re),  "=m" ((dest)->re)	\
	: "m" ((source1)->re), "m" ((source2)->re) )
/* Complex multiplication according to
   "Using Streaming SIMD Extensions 3 in Algorithms with Complex Arithmetic"
   Version 1.0 */
/* *** gcc-Version 3.3.1 assembler doesn't support addsupd and movddup
#define multiply(source1, source2, dest) \
__asm__ __volatile__ (			\
	"movddup %0, %%xmm0\n\t"	\
	"movapd %1, %%xmm1\n\t"		\
	"mulpd %%xmm1, %%xmm0\n\t"	\
	"shufpd $1, %%xmm1, %%xmm1\n\t"	\
	"movddup %3, %%xmm2\n\t"	\
	"mulpd %%xmm1, %%xmm2\n\t"	\
	"addsupd %%xmm2, %%xmm0\n\t"	\
	"movapd %%xmm0, %2"		\
	: "=m" ((source1)->re), "=m" ((source2)->re),  "=m" ((dest)->re), "=m" ((source1)->im)        \
	: "m" ((source1)->re), "m" ((source2)->re), "m" ((dest)->re), "m" ((source1)->im) )
*/
#define multiply(source1, source2, dest) \
__asm__ __volatile__ (			\
	"movapd %0, %%xmm0\n\t"		\
	"movapd %1, %%xmm1\n\t"		\
	"movapd %%xmm0, %%xmm2\n\t"	\
	"unpcklpd %%xmm0, %%xmm2\n\t"	\
	"mulpd %%xmm1, %%xmm2\n\t"	\
	"movapd %%xmm0, %%xmm3\n\t"	\
	"unpckhpd %%xmm0, %%xmm3\n\t"	\
	"mulpd %%xmm1, %%xmm3\n\t"	\
	"shufpd $1, %%xmm3, %%xmm3\n\t"	\
	"xorpd %%xmm7, %%xmm3\n\t"	\
	"addpd %%xmm3, %%xmm2\n\t"	\
	"movapd %%xmm2, %2"		\
	: "=m" ((source1)->re), "=m" ((source2)->re),  "=m" ((dest)->re), "=m" ((source1)->im)        \
	: "m" ((source1)->re), "m" ((source2)->re), "m" ((dest)->re), "m" ((source1)->im) )
#define multiply_add(source1, source2, dest) \
__asm__ __volatile__ (			\
	"movapd %0, %%xmm0\n\t"		\
	"movapd %1, %%xmm1\n\t"		\
	"movapd %2, %%xmm4\n\t"		\
	"movapd %%xmm0, %%xmm2\n\t"	\
	"unpcklpd %%xmm0, %%xmm2\n\t"	\
	"mulpd %%xmm1, %%xmm2\n\t"	\
	"movapd %%xmm0, %%xmm3\n\t"	\
	"unpckhpd %%xmm0, %%xmm3\n\t"	\
	"mulpd %%xmm1, %%xmm3\n\t"	\
	"shufpd $1, %%xmm3, %%xmm3\n\t"	\
	"xorpd %%xmm7, %%xmm3\n\t"	\
	"addpd %%xmm3, %%xmm2\n\t"	\
	"addpd %%xmm2, %%xmm4\n\t"	\
	"movapd %%xmm4, %2"		\
	: "=m" ((source1)->re), "=m" ((source2)->re),  "=m" ((dest)->re), "=m" ((source1)->im)        \
	: "m" ((source1)->re), "m" ((source2)->re), "m" ((dest)->re), "m" ((source1)->im) )
#define multiply_sub(source1, source2, dest) \
__asm__ __volatile__ (			\
	"movapd %0, %%xmm0\n\t"		\
	"movapd %1, %%xmm1\n\t"		\
	"movapd %2, %%xmm4\n\t"		\
	"movapd %%xmm0, %%xmm2\n\t"	\
	"unpcklpd %%xmm0, %%xmm2\n\t"	\
	"mulpd %%xmm1, %%xmm2\n\t"	\
	"movapd %%xmm0, %%xmm3\n\t"	\
	"unpckhpd %%xmm0, %%xmm3\n\t"	\
	"mulpd %%xmm1, %%xmm3\n\t"	\
	"shufpd $1, %%xmm3, %%xmm3\n\t"	\
	"xorpd %%xmm7, %%xmm3\n\t"	\
	"addpd %%xmm3, %%xmm2\n\t"	\
	"subpd %%xmm2, %%xmm4\n\t"	\
	"movapd %%xmm4, %2"		\
	: "=m" ((source1)->re), "=m" ((source2)->re),  "=m" ((dest)->re), "=m" ((source1)->im)        \
	: "m" ((source1)->re), "m" ((source2)->re), "m" ((dest)->re), "m" ((source1)->im) )
#else
#define add(source1, source2, dest) {\
   (dest)->re=(source1)->re + (source2)->re;\
   (dest)->im=(source1)->im + (source2)->im; }
#define sub(source1, source2, dest) {\
   (dest)->re=(source1)->re - (source2)->re;\
   (dest)->im=(source1)->im - (source2)->im; }
#define multiply(source1, source2, dest) {\
   (dest)->re=(source1)->re * (source2)->re - (source1)->im * (source2)->im;\
   (dest)->im=(source1)->im * (source2)->re + (source1)->re * (source2)->im; }
#define multiply_add(source1, source2, dest) {\
   multiply_add_re(source1, source2, (dest)->re); \
   multiply_add_im(source1, source2, (dest)->im); }
#define multiply_sub(source1, source2, dest) {\
   multiply_sub_re(source1, source2, (dest)->re); \
   multiply_sub_im(source1, source2, (dest)->im); }
#endif
#define complex_conjugate(source, dest) {\
   (dest)->re = (source)->re; \
   (dest)->im = -(source)->im; }
#else
  extern void add(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest);
  extern void sub(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest);
  extern void multiply(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest);
  extern void multiply_add(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest);
  extern void multiply_aub(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest);
  extern void complex_conjugate(COMPLEX *source, COMPLEX *dest);
#endif

extern void divide(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest);
/* Trigonometric and related functions */
extern REAL complex_abs_sq(COMPLEX *num);
extern REAL complex_abs(COMPLEX *num);
extern REAL complex_re(COMPLEX *num);
extern REAL complex_im(COMPLEX *num);
extern REAL complex_arg(COMPLEX *num);
extern void complex_sqrt(COMPLEX *source, COMPLEX *dest);
extern void complex_exp(COMPLEX *source, COMPLEX *dest);
extern void complex_ln(COMPLEX *source, COMPLEX *dest);
extern void complex_pow(COMPLEX *source1, COMPLEX *source2, COMPLEX *dest);
extern void complex_sin(COMPLEX *source, COMPLEX *dest);
extern void complex_cos(COMPLEX *source, COMPLEX *dest);
extern void complex_tan(COMPLEX *source, COMPLEX *dest);
extern void complex_cotan(COMPLEX *source, COMPLEX *dest);
extern void complex_sinh(COMPLEX *source, COMPLEX *dest);
extern void complex_cosh(COMPLEX *source, COMPLEX *dest);
extern void complex_tanh(COMPLEX *source, COMPLEX *dest);
extern void complex_cotanh(COMPLEX *source, COMPLEX *dest);
extern void complex_asinh(COMPLEX *source, COMPLEX *dest);
extern void complex_acosh(COMPLEX *source, COMPLEX *dest);
extern void complex_atanh(COMPLEX *source, COMPLEX *dest);
extern void complex_acotanh(COMPLEX *source, COMPLEX *dest);
extern void complex_asin(COMPLEX *source, COMPLEX *dest);
extern void complex_acos(COMPLEX *source, COMPLEX *dest);
extern void complex_atan(COMPLEX *source, COMPLEX *dest);
extern void complex_acotan(COMPLEX *source, COMPLEX *dest);
/* Outputting complex numbers */
extern void complex_to_str(COMPLEX *num, UBYTE *str);
extern void show_complex(COMPLEX *num);
extern void output_complex(COMPLEX *num, FILE *fp);
