/* Standard-Definitions */
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "variables.h"
#include "text.h"
#include "matrix.h"
#include "vectors.h"
#include "extrapolate.h"
#include "set.h"


/* *****************************
   Extrapolations-Verfahren */

/* das van den Broeck-Schwartz-Verfahren */

void extrapolate_vbs(struct vector *data, COMPLEX *alpha, COMPLEX *result)
 {
   struct vector approx[2], eps[2];
   int cnt;
   register int m, n;
   COMPLEX unity, temp0, temp1, temp2;
   REAL error;

   cnt = data->el_cnt;   
   alloc_int_vector(&(approx[0]), cnt);			/* Hilfsvektoren allokatieren */
   alloc_int_vector(&(approx[1]), cnt);
   alloc_int_vector(&(eps[0]), cnt);
   alloc_int_vector(&(eps[1]), cnt);

   copy_vector(data, &(approx[0]));			/* Daten kopieren */

   unity.re = 1.0;
   unity.im = 0.0;

   for(m=1; m<(cnt+1)/2; m++)				/* Approximanten bilden */
    {
      for(n=m-1; n<cnt-m; n++)
       {						/* Epsilon-Folgen durchgehen */
         sub(&(approx[(m-1) & 1].elements[n+1]), &(approx[(m-1) & 1].elements[n]), &temp0);
         divide(&unity, &temp0, &temp1);
         multiply(alpha, &(eps[(m-1) & 1].elements[n]), &temp2);
         add(&temp1, &temp2, &(eps[m & 1].elements[n]));
       }
      for(n=m; n<cnt-m; n++)
       {						/* Approximanten-Folgen durchgehen */
         sub(&(eps[m & 1].elements[n]), &(eps[m & 1].elements[n-1]), &temp0);
         divide(&unity, &temp0, &temp1);
         add(&(approx[(m-1) & 1].elements[n]), &temp1, &(approx[m & 1].elements[n]));

/*         printf("[%d,%d] = ", n, m);
         show_complex(&(approx[m & 1].elements[n])); */
       }
    }

   result->re = approx[((cnt-1)/2) & 1].elements[cnt/2].re;
   result->im = approx[((cnt-1)/2) & 1].elements[cnt/2].im;

   sub(&(approx[(((cnt-1)/2)-1) & 1].elements[cnt/2]), result, &temp0);		/* Fehler schaetzen */
   sub(&(approx[(((cnt-1)/2)-1) & 1].elements[(cnt/2)+1]), result, &temp1);
   error = sqrt((complex_abs_sq(&temp0)+complex_abs_sq(&temp1)) / 2.0);
   temp0.re = error;								/* Fehler speichern */
   temp0.im = 0;
   assign_variable((UBYTE *) "error_vbs", &temp0);

   free_vector(&(eps[1]));				/* Hilfsvektoren freigeben */
   free_vector(&(eps[0]));
   free_vector(&(approx[1]));
   free_vector(&(approx[0]));
 }

/* das Stoer-Burlisch-Verfahren */

void extrapolate_stb(struct vector *data, struct vector *h, COMPLEX *gamma, COMPLEX *result)
 {
   struct vector approx[3];
   int cnt;
   register int m, n;
   COMPLEX unity, temp0, temp1, temp2, temp3;
   REAL error;

   cnt = data->el_cnt;
   if(cnt != h->el_cnt)
    {
      display_error("Unequal data- and argument-count in Stoer-Burlisch extrapolation");
      return;
    }

   alloc_int_vector(&(approx[0]), cnt);			/* Hilfsvektoren allokatieren */
   alloc_int_vector(&(approx[1]), cnt);
   alloc_int_vector(&(approx[2]), cnt);

   copy_vector(data, &(approx[0]));			/* Daten kopieren */

   unity.re = 1.0;
   unity.im = 0.0;

   for(m=1; m<cnt; m++)					/* Approximanten bilden */
    {
      for(n=0; n<cnt-m; n++)
       {						/* Approximanten-Folgen durchgehen */
         sub(&(approx[(m-1) % 3].elements[n+1]), &(approx[(m-1) % 3].elements[n]), &temp0);
         if(m > 1)
          {
           sub(&(approx[(m-1) % 3].elements[n+1]), &(approx[(m-2) % 3].elements[n+1]), &temp1);
          }
         else
          {
            temp1.re = approx[(m-1) % 3].elements[n+1].re;
            temp1.im = approx[(m-1) % 3].elements[n+1].im;
          }
         if(complex_abs_sq(&temp1))		/* Achtung, falls Folge konstant */
           divide(&temp0, &temp1, &temp2);
         else
           temp2.re = temp2.im = 0.0;
         sub(&unity, &temp2, &temp3);
         divide(&(h->elements[n]), &(h->elements[n+m]), &temp1);
         complex_pow(&temp1, gamma, &temp2);
         multiply(&temp2, &temp3, &temp1);
         sub(&temp1, &unity, &temp2);
         if(complex_abs_sq(&temp2))		/* Achtung, falls Folge konstant */
            divide(&temp0, &temp2, &temp3);
         else
            temp3.re = temp3.im = 0.0;
         add(&(approx[(m-1) % 3].elements[n+1]), &temp3, &(approx[m % 3].elements[n]));

/*         printf("[%d,%d] = ", n, m);
         show_complex(&(approx[m % 3].elements[n])); */
       }
    }

   result->re = approx[(cnt-1) % 3].elements[0].re;
   result->im = approx[(cnt-1) % 3].elements[0].im;

   sub(&(approx[(cnt-2) % 3].elements[0]), result, &temp0);		/* Fehler schaetzen */
   sub(&(approx[(cnt-2) % 3].elements[1]), result, &temp1);
   error = sqrt((complex_abs_sq(&temp0)+complex_abs_sq(&temp1)) / 2.0);
   temp0.re = error;								/* Fehler speichern */
   temp0.im = 0;
   assign_variable((UBYTE *) "error_stb", &temp0);

   free_vector(&(approx[0]));				/* Hilfsvektoren freigeben */
   free_vector(&(approx[1]));
   free_vector(&(approx[2]));
 }

/* Polynomiale Interpolation */

void interpolate_poly(struct vector *data, struct vector *h, COMPLEX *x, COMPLEX *result)
 {
   int cnt;
   register int i, j;
   COMPLEX temp0, temp1, temp2, temp3;

   cnt = data->el_cnt;
   if(cnt != h->el_cnt)
    {
      display_error("Unequal data- and argument-count in polynomial interpolation");
      return;
    }

   result->re = result->im = 0.0;

   for(i=0; i<cnt; i++)
    {
      temp0.re = 1.0;
      temp0.im = 0.0;
      for(j=0; j<cnt; j++)
        if(j != i)
          {
           sub(x, &(h->elements[j]), &temp1);
           sub(&(h->elements[i]), &(h->elements[j]), &temp2);
           multiply(&temp0, &temp1, &temp3);
           divide(&temp3, &temp2, &temp0);
          }
      multiply(&temp0, &(data->elements[i]), &temp1);
      add(&temp1, result, result);
    }
 }

/* Lineare Regression */

void lin_reg(struct vector *x, struct vector *y, COMPLEX *a, COMPLEX *b, COMPLEX *result)
 {
   int cnt, cntp;
   register int i;
   COMPLEX xb, yb, xsqb, xy;
   COMPLEX temp0, temp1, temp2, temp3;

   cnt = x->el_cnt;
   cntp = cnt-STARTFIT+1;
   if(cnt != y->el_cnt)
    {
      display_error("Unequal number of x and y points in linear regression");
      return;
    }

  xb.re = xb.im = 0;
  yb.re = yb.im = 0;
  xsqb.re = xsqb.im = 0;
  xy.re = xy.im = 0;
  for(i=STARTFIT-1; i<cnt; i++)
   {
    add(&(x->elements[i]), &xb, &xb);
    add(&(y->elements[i]), &yb, &yb);
    multiply(&(x->elements[i]), &(x->elements[i]), &temp0);
    add(&temp0, &xsqb, &xsqb);
    multiply(&(x->elements[i]), &(y->elements[i]), &temp0);
    add(&temp0, &xy, &xy);
   }
  multiply(&xb, &xb, &temp0);
  temp0.re /= cntp;
  temp0.im /= cntp;
  sub(&xsqb, &temp0, &temp1);

  multiply(&xb, &yb, &temp2);
  temp2.re /= cntp;
  temp2.im /= cntp;
  sub(&xy, &temp2, &temp3);

  divide(&temp3, &temp1, b);
  multiply(b, &xb, &temp0);
  sub(&yb, &temp0, a);
  a->re /= cntp;
  a->im /= cntp;

  temp0.re = temp0.im = 0;
  for(i=STARTFIT-1; i<cnt; i++)
   {
    multiply(b, &(x->elements[i]), &temp1);
    add(&temp1, a, &temp2);
    sub(&(y->elements[i]), &temp2, &temp3);
    multiply(&temp3, &temp3, &temp2);
    add(&temp2, &temp0, &temp0);
   }
  temp0.re /= cntp;
  temp0.im /= cntp;

  complex_sqrt(&temp0, result);
 }

/* Finite Fourier Transformation */

void finite_fourier(struct vector *data, struct vector *si, struct vector *co, COMPLEX *result)
 {
   int d, dd;
   register int i, j;

   d = data->el_cnt;
   dd= d/2;
   if((dd != si->el_cnt) && (dd != co->el_cnt))
    {
      display_error("The vectors for the coefficients of sin and cos have inappropriate length");
      return;
    }

   result->re = result->im = 0;
   for(i=0; i<d; i++)
    {
      result->re += data->elements[i].re;
      result->im += data->elements[i].im;
    }

   result->re /= d;
   result->im /= d;

   for(j=0; j<dd; j++)
    {
     si->elements[j].re = si->elements[j].im = 0;
     for(i=0; i<d; i++)
      {
        si->elements[j].re += data->elements[i].re*sin((2.0*PI*(j+1)*i)/d);
        si->elements[j].im += data->elements[i].im*sin((2.0*PI*(j+1)*i)/d);
      }
     si->elements[j].re *= 2.0;
     si->elements[j].im *= 2.0;
     si->elements[j].re /= d;
     si->elements[j].im /= d;
     co->elements[j].re = co->elements[j].im = 0;
     for(i=0; i<d; i++)
      {
        co->elements[j].re += data->elements[i].re*cos((2.0*PI*(j+1)*i)/d);
        co->elements[j].im += data->elements[i].im*cos((2.0*PI*(j+1)*i)/d);
      }
     co->elements[j].re *= 2.0;
     co->elements[j].im *= 2.0;
     co->elements[j].re /= d;
     co->elements[j].im /= d;
    }
 }

/* Complex Finite Fourier Transformation */


void complex_finite_fourier(struct vector *data, struct vector *four, COMPLEX *result)
 {
   int d;
   register int i, j;

   d = data->el_cnt;
   if(d != four->el_cnt)
    {
      display_error("The vectors for the coefficients of the fourier transform have inappropriate length");
      return;
    }

   for(j=0; j<d; j++)
    {
     four->elements[j].re = four->elements[j].im = 0;
     for(i=0; i<d; i++)
      {
        four->elements[j].re += (data->elements[i].re*cos((2.0*PI*j*i)/d) - data->elements[i].im*sin((2.0*PI*j*i)/d));
        four->elements[j].im += (data->elements[i].im*cos((2.0*PI*j*i)/d) + data->elements[i].re*sin((2.0*PI*j*i)/d));
      }
     four->elements[j].re /= d;
     four->elements[j].im /= d;
    }

   result->re =  four->elements[0].re;
   result->im =  four->elements[0].im;
 }

