/* *** 18.12.2003:
      - output_Nexpr() (command "output"):
        file buffers are kept open (should speed up writing files, at
        least under some conditions)
*/

/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "evaluate.h"
#include "variables.h"
#include "text.h"
#include "matrix.h"
#include "vectors.h"
#include "extrapolate.h"
#include "set.h"
#include "hamilton.h"

/* ***********************************************************
   Subroutines related to Z_n Quantum-Spin-chains */

/* Alle Files eines Z_4-Datensatzes lesen */

#define SITEDIFF (END_SITES-START_SITES+1)

void read_z4(UBYTE *ident, UBYTE *fname)
 {
   static UBYTE hstring[256];
   register int i, n, q;
   COMPLEX temp1, temp2, temp3, temp4, temp5, temp6;
   COMPLEX result;
   struct vector *d, *h;

   for(i=0; i<LEVELS; i++)
    for(q = 0; q<CHARGE; q++)
      {
        sprintf(hstring, "d%dQ%d", i, q);
        alloc_vector(hstring, SITEDIFF);
      }
   alloc_vector((UBYTE *) "h", SITEDIFF);

   for(n = START_SITES; n<=END_SITES; n++)
    {
     sprintf(hstring, "h(%d) := 1/%d;", n-START_SITES+1, n);
     interpret_line(hstring, &result, FALSE);
     for(q = 0; q<CHARGE; q++)
      {
        sprintf(hstring, "%sh%dQ%d%sout", fname, n, q, ident);
        read_file(hstring, FALSE);
        for(i=0; i<LEVELS; i++)
         {
           if(! q)
             sprintf(hstring, "d%dQ%d(%d) := Q0e%d - Q0e0;", i, q, n-START_SITES+1, i+1);
           else
             sprintf(hstring, "d%dQ%d(%d) := e%d - Q0e0;", i, q, n-START_SITES+1, i);
           interpret_line(hstring, &result, FALSE);
         }
      }
    }

   /* Die y's berechnen */

   for(i=0; i<LEVELS; i++)
    for(q = 0; q<CHARGE; q++)
      {
        sprintf(hstring, "y%dQ%d", i, q);
        alloc_vector(hstring, SITEDIFF-1);
      }
   alloc_vector((UBYTE *) "hy", SITEDIFF-1);
   for(n = START_SITES+1; n<=END_SITES; n++)
    {
     sprintf(hstring, "hy(%d) := 1/%d;", n-START_SITES, n);
     interpret_line(hstring, &result, FALSE);
    }
   for(i=0; i<LEVELS; i++)
    for(q = 0; q<CHARGE; q++)
      {
        sprintf(hstring, "d%dQ%d", i, q);
        d = identify_vector(hstring, TRUE);
        sprintf(hstring, "e%dQ%d", i, q);
        interpret_line(hstring, &temp1, FALSE);
        for(n=START_SITES+1; n<=END_SITES; n++)
         {
           sub(&(d->elements[n-START_SITES]), &temp1, &temp2);
           sub(&(d->elements[n-START_SITES-1]), &temp1, &temp3);
           divide(&temp2, &temp3, &temp4);
           complex_ln(&temp4, &temp5);
           temp2.re = n;
           temp2.im = 0.0;
           temp3.re = n-1;
           temp3.im = 0.0;
           divide(&temp2, &temp3, &temp4);
           complex_ln(&temp4, &temp6);
           divide(&temp5, &temp6, &temp2);
           sprintf(hstring, "y%dQ%d(%d) := %17.17g+i*(%17.17g)", i, q, n-START_SITES, temp2.re, temp2.im);
           interpret_line(hstring, &result, FALSE);
         }
      }

   /* Extrapolation der Energieluecken durchfuehren */
   h = identify_vector((UBYTE *) "h", TRUE);
   temp1.re = -1.0;
   temp1.im = 0.0;
   temp2.re = 0.8;
   temp2.im = 0.0;
   for(q = 0; q<CHARGE; q++)
    {
      for(i=0; i<LEVELS; i++)
        {
           sprintf(hstring, "d%dQ%d", i, q);
           d = identify_vector(hstring, TRUE);
           extrapolate_vbs(d, &temp1, &result);
           printf("Estimate for d%dQ%d by vbs: ", i, q);
           show_complex(&result);
           sprintf(hstring, "e%dQ%d", i, q);
           assign_variable(hstring, &result);
           evaluate((UBYTE *) "error_vbs", &result);
           printf("  Estimated error: ");
           show_complex(&result);
           extrapolate_stb(d, h, &temp2, &result);
           printf("Estimate for d%dQ%d by stb: ", i, q);
           show_complex(&result);
           evaluate((UBYTE *) "error_stb", &result);
           printf("  Estimated error: ");
           show_complex(&result);
        }
    }

   /* Extrapolation der Exponenten durchfuehren */
   h = identify_vector((UBYTE *) "hy", TRUE);
   temp1.re = -1.0;
   temp1.im = 0.0;
   temp2.re = 0.8;
   temp2.im = 0.0;
   for(q = 0; q<CHARGE; q++)
    {
      for(i=0; i<LEVELS; i++)
        {
           sprintf(hstring, "y%dQ%d", i, q);
           d = identify_vector(hstring, TRUE);
           extrapolate_vbs(d, &temp1, &result);
           printf("Estimate for y%dQ%d by vbs: ", i, q);
           show_complex(&result);
           sprintf(hstring, "y%dQ%d", i, q);
           assign_variable(hstring, &result);
           evaluate((UBYTE *) "error_vbs", &result);
           printf("  Estimated error: ");
           show_complex(&result);
           extrapolate_stb(d, h, &temp2, &result);
           printf("Estimate for y%dQ%d by stb: ", i, q);
           show_complex(&result);
           evaluate((UBYTE *) "error_stb", &result);
           printf("  Estimated error: ");
           show_complex(&result);
        }
    }
 }

/* Alle Files eines Z_3-Impuls-Datensatzes lesen */

#define START_IMPULSE 0
#define END_IMPULSE (SITESP-1)

void read_z3p(UBYTE *ident, UBYTE *fname)
 {
   static UBYTE hstring[256];
   register int i, q, p, q1, p1;
   COMPLEX result;
   COMPLEX **low_levels;

   low_levels = (COMPLEX **) malloc(CHARGE1 * sizeof(COMPLEX *));
   if(! low_levels)
     goto oomem;
   for(i=0; i<CHARGE1; i++)
    {
      low_levels[i] = (COMPLEX *) malign16(SITESP * sizeof(COMPLEX));
      if(! low_levels[i])
       {
         for(q=0; q<i; q++)
            free(low_levels[i]);
         free(low_levels);
         goto oomem;
       }
    }   

   for(p=START_IMPULSE; p<=END_IMPULSE; p++)
    for(q = 0; q<CHARGE1; q++)
      {
        sprintf(hstring, "dQ%dp%d", q, p);
        alloc_vector(hstring, LEVELS1);
      }

   for(p = START_IMPULSE; p<=END_IMPULSE; p++)
    {
     for(q = 0; q<CHARGE1; q++)
      {
        sprintf(hstring, "%sQ%dp%d%sout", fname, q, p, ident);
        read_file(hstring, FALSE);
        if((! q) && (! p))
           interpret_line("Q0e0 := e0", &result, FALSE);
        for(i=0; i<LEVELS1; i++)
         {
           sprintf(hstring, "dQ%dp%d(%d) := e%d - Q0e0;", q, p, i+1, ((! q) && (! p))?i+1:i);
           interpret_line(hstring, &result, FALSE);
         }
      }
    }

   for(q=0; q<CHARGE1; q++)				/* Levels initialisieren */
    for(p = START_IMPULSE; p<=END_IMPULSE; p++)
     {
       (low_levels[q]+p)->re = 10000000.0;
       (low_levels[q]+p)->im = 0.0;
     }

   for(q=1; q<CHARGE1; q++)				/* Niedrigste Zweiteilchen-Levels suchen */
    for(q1=q; q1<CHARGE1; q1++)
      for(p = START_IMPULSE; p<=END_IMPULSE; p++)
        for(p1 = p; p1<=END_IMPULSE; p1++)
          {
            sprintf(hstring, "dQ%dp%d(1) + dQ%dp%d(1)", q, p, q1, p1);
            interpret_line(hstring, &result, FALSE);
            if(result.re < (low_levels[(q+q1) % CHARGE1] + ((p+p1) % (END_IMPULSE+1)))->re)
              {
                  (low_levels[(q+q1) % CHARGE1] + ((p+p1) % (END_IMPULSE+1)))->re = result.re;
                  (low_levels[(q+q1) % CHARGE1] + ((p+p1) % (END_IMPULSE+1)))->im = result.im;
              }
          }

   for(q=0; q<CHARGE1; q++)				/* Levels ausgeben */
    for(p = START_IMPULSE; p<=END_IMPULSE; p++)
     {
       printf("Lowest two particle level for Q=%d, p=%d is: ",q ,p);
       show_complex(low_levels[q]+p);
     }

   for(i=0; i<CHARGE1; i++)
      free(low_levels[i]);
   free(low_levels);
   return;
 oomem:
   display_error("Out of memory while reading impulse spectrum");
 }

/* Eine Tabelle erzeugen */

void write_table(UBYTE *fname)
 {
   FILE *fpo;
   register int q, i, n;
   int t;
   struct vector *vec;
   UBYTE hstring[200];

   fname = strip_file(fname);
   fpo = fopen(fname, "w");
   if(! fpo)
    {
      display_error_text("Failed to write '%s'", fname);
      return;
    }

   for(q=0; q<CHARGE; q++)
    {
      fprintf(fpo, "$Q = %d$\n", q);
      fprintf(fpo, "& $ N$ &");
      for(i=0; i<LEVELS; i++)
       {
         if(! q)
           t = i+1;
         else
           t = i;
         fprintf(fpo, "&  $ i = %d $   &", t);
       }
      fprintf(fpo, " \\cr \\tablerule\n");
      for(n=START_SITES; n<=END_SITES; n++)
       {
         fprintf(fpo, "& $%2d$ &", n);
         for(i=0; i<LEVELS; i++)
          {
            sprintf(hstring, "d%dQ%d", i, q);
            vec = identify_vector(hstring, TRUE);
            if(vec)
              fprintf(fpo, "& $%10.8g$ &", vec->elements[n-START_SITES].re);
          }
         fprintf(fpo, " \\cr \\tablespace\n");
       }
      fprintf(fpo, "\n");
    }

   fclose(fpo);
 }

/* Ein Gnu-Plotfile fuer Z_4-Daten erzeugen */

void write_plot(UBYTE *fname)
 {
   FILE *fpo;
   register int q, i, n;
   int t;
   struct vector *vec;
   UBYTE hstring[200];

   fname=strip_file(fname);
   for(q=0; q<CHARGE; q++)
    {
      sprintf(hstring, "%s%d", fname, q);
      fpo = fopen(hstring, "w");
      if(! fpo)
       {
         display_error_text("Failed to write '%s'", hstring);
         return;
       }

      for(n=START_SITES; n<=END_SITES; n++)
       {
         for(i=0; i<LEVELS; i++)
          {
            fprintf(fpo, "%d ", n);
            sprintf(hstring, "d%dQ%d", i, q);
            vec = identify_vector(hstring, TRUE);
            if(vec)
              fprintf(fpo, "%10.8g\n", vec->elements[n-START_SITES].re);
          }
       }
     fclose(fpo);
    }
 }

/* Ein Gnu-Plotfile fuer Z_3-Impuls-Daten erzeugen */

void write_z3_plot(UBYTE *fname)
 {
   FILE *fpo;
   register int q, i, p;
   int t;
   struct vector *vec;
   UBYTE hstring[200];

   fname=strip_file(fname);
   for(q=0; q<CHARGE1; q++)
    {
      sprintf(hstring, "%s%d", fname, q);
      fpo = fopen(hstring, "w");
      if(! fpo)
       {
         display_error_text("Failed to write '%s'", hstring);
         return;
       }

      for(p=START_IMPULSE; p<=END_IMPULSE; p++)
       {
         sprintf(hstring, "dQ%dp%d", q, p);
         vec = identify_vector(hstring, TRUE);
         if(vec)
          for(i=0; i<LEVELS1; i++)
           {
            fprintf(fpo, "%d ", p);
            fprintf(fpo, "%10.8g\n", vec->elements[i].re);
           }
       }
     fclose(fpo);
    }
 }

/* Ein Gnu-Plotfile fuer Impuls-Daten erzeugen */
/* Konventionen P=0,2pi/N,...,(N-1) 2pi/N */

void write_P_plot(UBYTE *fname)
 {
   FILE *fpo;
   register int q, i, p;
   int t;
   struct vector *vec;
   UBYTE hstring[200];

   fname=strip_file(fname);
   for(q=0; q<CHARGE1; q++)
    {
      sprintf(hstring, "%s%d", fname, q);
      fpo = fopen(hstring, "w");
      if(! fpo)
       {
         display_error_text("Failed to write '%s'", hstring);
         return;
       }

      for(p=START_IMPULSE; p<=END_IMPULSE; p++)
       {
         sprintf(hstring, "dQ%dp%d", q, p);
         vec = identify_vector(hstring, TRUE);
         if(vec)
          for(i=0; i<LEVELS1; i++)
           {
            if(! p)
                fprintf(fpo,"0     ");
            else
                fprintf(fpo, "%10.8g ", 2.0*PI/SITESP*(SITESP-p));
            fprintf(fpo, "%10.8g\n", vec->elements[i].re);
           }
       }
     fclose(fpo);
    }
 }

/* Zwei Vektoren in ein Gnuplot-File schreiben */

void output_plot(UBYTE *vec1, UBYTE *vec2, UBYTE *name)
 {
   FILE *fpo;
   register int i;
   struct vector *veca, *vecb;
   char format_str[64];

   for(i=0; name[i]; i++)
     if((name[i] == ';') || (name[i] == '$'))
        name[i] = 0;
   name = strip_file(name);

   veca=identify_vector(vec1, TRUE);
   vecb=identify_vector(vec2, TRUE);

   if(veca && vecb)
    {
      if(veca->el_cnt != vecb->el_cnt)
       {
         display_error("The vectors in outplot have unequal length");
         return;
       }
      fpo = fopen(name, "w");
      if(! fpo)
       {
         display_error_text("Failed to write '%s'", name);
         return;
       }
      /* Define format string */
      sprintf(format_str, "%%%d.%dg", OUTDIGITS+2, OUTDIGITS);

      for(i=0; i<veca->el_cnt; i++)
       {
        fprintf(fpo, format_str, veca->elements[i].re);
        fprintf(fpo, "\t");
        fprintf(fpo, format_str, vecb->elements[i].re);
        fprintf(fpo, "\n");
       }
      fclose(fpo);
    }
 }

/* Einen Vektor aus einem File mit Indizes lesen */

void read_ind(UBYTE *vecn, UBYTE *name)
 {
   FILE *fpi;
   UBYTE temp_str[200];
   register int i;
   int ind; 
   struct vector *vec;
   REAL hval;

   for(i=0; name[i]; i++)
     if((name[i] == ';') || (name[i] == '$'))
        name[i] = 0;
   name = strip_file(name);

   vec=identify_vector(vecn, TRUE);

   if(vec)
    {
      fpi = fopen(name, "r");
      if(! fpi)
       {
         display_error_text("Failed to open '%s' for reading", name);
         return;
       }
      while(fgets(temp_str, 200, fpi))
       if((*temp_str != '%') && (*temp_str != '#'))
       {
        sscanf(temp_str, "%d", &ind);
        ind--;
        if((ind >= 0) && (ind < vec->el_cnt))
         {
          for(i=0; (temp_str[i] >= '0') && (temp_str[i] <= '9'); i++)
                   ;
          while(temp_str[i] && (temp_str[i] <= ' '))
                     i++;
          sscanf(temp_str+i, "%lg", &hval);
          vec->elements[ind].re = hval;
          vec->elements[ind].im = 0;
         }
        else
          display_error("Index in wrong range");
       }
      fclose(fpi);
    }
 }

/* Zwei Vektoren aus einer Datei lesen */

void read_vecs(UBYTE *vecx, UBYTE *vecy, UBYTE *name)
 {
   FILE *fpi;
   UBYTE temp_str[200];
   register int i;
   int ind;
   struct vector *veca, *vecb;
   REAL hval;
 
   for(i=0; name[i]; i++)
     if((name[i] == ';') || (name[i] == '$'))
        name[i] = 0;
   name = strip_file(name);
 
   veca=identify_vector(vecx, TRUE);
   vecb=identify_vector(vecy, TRUE);
 
   if(veca && vecb)
    {
      fpi = fopen(name, "r");
      if(! fpi)
       {
         display_error_text("Failed to open '%s' for reading", name);
         return;
       }
      ind = 0;
      while(fgets(temp_str, 200, fpi))
       if((*temp_str != '%') && (*temp_str != '#'))
       {
        if((ind >= 0) && (ind < veca->el_cnt) && (ind < vecb->el_cnt))
         {
          i=0;
          while(temp_str[i] && (temp_str[i] <= ' '))
                     i++;
          if((! (temp_str[i])) || (temp_str[i] == '\n'))
            goto loop;
          sscanf(temp_str+i, "%lg", &hval);
          veca->elements[ind].re = hval;
          veca->elements[ind].im = 0;
          while(temp_str[i] > ' ')
                   i++;
          while(temp_str[i] && (temp_str[i] <= ' '))
                     i++;
          sscanf(temp_str+i, "%lg", &hval);
          vecb->elements[ind].re = hval;
          vecb->elements[ind].im = 0;
         }
        else
          display_error("Index in wrong range");
        ind++;
     loop: ;
       }
      fclose(fpi);
    }
 }

/* Buffers for output */

static FILE *output_fpo[MAXOUT];
static char *output_name[MAXOUT];
static int output_cnt = 0;

/* Close output buffers */

void output_close()
 {
  register int i;

  for(i=0; i<output_cnt; i++)
   {
    if(output_fpo[i])
      fclose(output_fpo[i]);
    free(output_name[i]);
   }
 }

/* Output N Expressions */

void output_Nexpr(int cnt, char *exprN[])
 {
   FILE *fpo;
   register int j, i;
   COMPLEX ex[MAXARGOUT];
   char *name;
   char format_str[64];

   name = exprN[--cnt];

   for(i=0; name[i]; i++)
     if((name[i] == ';') || (name[i] == '$'))
        name[i] = 0;
   name = strip_file(name);

   for(j=0; j<cnt; j++)
    {
     for(i=0; exprN[j][i] > ' '; i++)
              ;
     exprN[j][i] = 0;
     interpret_line(exprN[j], &(ex[j]), FALSE);
    }

   for(i=0; (i<output_cnt) && strcmp(name, output_name[i]); i++)
       ;
   if(i >= output_cnt)
    {
     if(output_cnt >= MAXOUT)
       {
        display_error_text("Cannot open '%s' for writing (too many files already open)", name);
        return;
       }
     if(! (output_name[i] = malloc(strlen(name)+1)))
       {
        display_error_text("Out of memory while opening '%s' for writing", name);
        return;
       }
     for(j=0; output_name[i][j] = name[j]; j++)
                 ;
     output_fpo[i] = fopen(name, "a");
     output_cnt++;
    }

   fpo = output_fpo[i];

   if(! fpo)
    {
      display_error_text("Failed to open '%s' for writing", name);
      return;
    }

   /* Define format string */
   sprintf(format_str, "%%%d.%dg", OUTDIGITS+2, OUTDIGITS);

   for(j=0; j<cnt; j++)
    {
     fprintf(fpo, format_str, ex[j].re);
     if(j<cnt-1)
        fprintf(fpo, "\t");
     else
        fprintf(fpo, "\n");
    }
   if(OUTFLUSH)
     fflush(fpo);
 }
