/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "variables.h"
#include "matrix.h"
#include "vectors.h"
#include "text.h"
#include "set.h"
#include "extrapolate.h"
#include "evaluate.h"
#include "integrate.h"

/* Variable, die ungleich Null sein sollte, wenn minmaximize() zeigen soll, was es tut */

#define minoutput 0


/* *******************************************
   Integrations-Funktionen */

/* Simpson-Integration */

void integrate(UBYTE *kernel, UBYTE *var, COMPLEX *lbd, COMPLEX *ubd, int open_flag, int intticks, COMPLEX *result)
 {
   COMPLEX int_len, h, sum, x;
   COMPLEX a, b, c;
   register int i;

/*   printf("Integrate '%s' with respect to '%s'\n", kernel, var);
   printf("from: ");
   show_complex(lbd);
   printf("to:   ");
   show_complex(ubd);
   if(open_flag)
      printf("in the open interval\n");
   else
      printf("in the closed interval\n");
   printf("with %d ticks\n", intticks); */

   sub(ubd, lbd, &int_len);
   if(open_flag)
    {
      a.re = lbd->re + int_len.re/(2*intticks+2);
      a.im = lbd->im + int_len.im/(2*intticks+2);
      b.re = ubd->re - int_len.re/(2*intticks+2);
      b.im = ubd->im - int_len.im/(2*intticks+2);
      integrate(kernel, var, &a, &b, FALSE, intticks, result);
      return;
    }

   h.re = int_len.re/(2*intticks);
   h.im = int_len.im/(2*intticks);
   x.re = lbd->re;
   x.im = lbd->im;

   assign_variable(var, lbd);
   evaluate(kernel, &sum);

   for(i=0; i<intticks; i++)
    {
     add(&x, &h, &a);
     assign_variable(var, &a);
     evaluate(kernel, &b);
     add(&a, &h, &c);
     assign_variable(var, &c);
     evaluate(kernel, &a);
     sum.re += (2*a.re + 4*b.re);
     sum.im += (2*a.im + 4*b.im);
     x.re += 2.0*h.re;
     x.im += 2.0*h.im;
    }

   assign_variable(var, ubd);
   evaluate(kernel, &a);

   sum.re = (sum.re-a.re)/3;
   sum.im = (sum.im-a.im)/3;

   multiply(&sum, &h, result);
 }

/* Integrieren und dann mit Stoer-Burlisch-Algorithmus extrapolieren */

#define SAMPLES 10

void integrate_extrapolate(UBYTE *kernel, UBYTE *var, COMPLEX *lbd, COMPLEX *ubd, int open_flag, int intticks, COMPLEX *result)
 {
   struct vector h, integral;
   register int i;
   COMPLEX gamma;

   alloc_int_vector(&h, SAMPLES);
   alloc_int_vector(&integral, SAMPLES);

   if(h.el_cnt && integral.el_cnt)
    {
      for(i=0; i<SAMPLES; i++)
       {
         h.elements[i].re = 1.0/(i+1);
         h.elements[i].im = 0.0;
         integrate(kernel, var, lbd, ubd, open_flag, intticks*(i+1)/SAMPLES, integral.elements+i);
       }
      gamma.re = 0.5;
      gamma.im = 0.0;
      extrapolate_stb(&integral, &h, &gamma, result);
    }

   free_vector(&h);
   free_vector(&integral);
 }

/* Das Cauchy-Integal um z0 mit Radius r auswerten */

static UBYTE c_var[] = "(_ci_r*exp(2*pi*i*_ci_t)+_ci_z0)";
static UBYTE c_subs[] = "*exp(2*pi*i*_ci_t)";

void integrate_cauchy(UBYTE *kernel, UBYTE *var, COMPLEX *z0, COMPLEX *rad, int intticks, COMPLEX *result)
 {
   UBYTE c_kernel[300];
   register int i, j;
   int flag, ofs, inccnt;
   COMPLEX temp_result;
   COMPLEX lbd, ubd;

   assign_variable((UBYTE *) "_ci_r", rad);	/* Radius speichern */
   assign_variable((UBYTE *)"_ci_z0", z0);	/* Mittelpunkt speichern */

   ofs=0;			/* Substitution durchfuehren */
   inccnt=strlen(var)-1;
   for(i=0; kernel[i]; i++)
    {
      flag=TRUE;
      if(i)
       if(((kernel[i-1] | 0x20) >= 'a') && ((kernel[i-1] | 0x20) <= 'z'))
          flag=FALSE;
      if(flag)
        {
         for(j=0; var[j] && flag; j++)
           if(conv_to_lower(kernel[i+j]) != conv_to_lower(var[j]))
                 flag = FALSE;
         if(flag)
           if(((kernel[i+j] | 0x20) >= 'a') && ((kernel[i+j] | 0x20) <= 'z'))
                 flag = FALSE;
        }
      if(! flag)
        c_kernel[ofs++] = kernel[i];
      else
       {
        for(j=0; c_var[j]; j++)
           c_kernel[ofs++] = c_var[j];
        i += inccnt;
       }
    }
   for(j=0; c_subs[j]; j++)
     c_kernel[ofs++] = c_subs[j];
   c_kernel[ofs] = 0;

/*   printf("int_cauchy(%s,%s,z0,r)\n", kernel, var);
   printf("  = int(%s,_ci_t,0,1)\n", c_kernel); */
   lbd.re = lbd.im = ubd.im = 0.0;
   ubd.re = 1.0;
   integrate(c_kernel, (UBYTE *) "_ci_t", &lbd, &ubd, FALSE, intticks, &temp_result);
   multiply(&temp_result, rad, result);
 }

/* *******************************************
   Differentiations-Funktionen */

/* Differentiation ueber Extrapolation mit Stoer-Burlisch */
/* In eine bestimmte Richtung, nach einem reellen Parameter */

void diff(UBYTE *kernel, UBYTE *var, COMPLEX *z0, COMPLEX *direc, int open_flag, int diffticks, COMPLEX *result)
 {
   struct vector h, tangent;
   register int i;
   COMPLEX gamma, dist;
   REAL temp_r;
   COMPLEX left, right, left_f, right_f;
   COMPLEX temp, temp1;

   temp_r = complex_abs(direc);
   dist.re = direc->re / temp_r;
   dist.im = direc->im / temp_r;

   alloc_int_vector(&h, diffticks);
   alloc_int_vector(&tangent, diffticks);

   if(h.el_cnt && tangent.el_cnt)
    {
      for(i=0; i<diffticks; i++)
       {
         h.elements[i].re = dist.re;
         h.elements[i].im = dist.im;
         right.re = z0->re + dist.re;
         right.im = z0->im + dist.im;
         assign_variable(var, &right);
         evaluate(kernel, &right_f);
         if(! open_flag)
          {
            left.re = z0->re - dist.re;
            left.im = z0->im - dist.im;
          }
         else
          {
            left.re = z0->re + dist.re/2.0;
            left.im = z0->im + dist.im/2.0;
          }
         assign_variable(var, &left);
         evaluate(kernel, &left_f);
         sub(&right_f, &left_f, &temp);
         if(! open_flag)
          {
            temp1.re = dist.re * 2.0;
            temp1.im = dist.im * 2.0;
          }
         else
          {
            temp1.re = dist.re / 2.0;
            temp1.im = dist.im / 2.0;
          }
         divide(&temp, &temp1, tangent.elements+i);
         dist.re /= 2.0;		/* Abstand halbieren */
         dist.im /= 2.0;
       }
      gamma.re = 0.5;
      gamma.im = 0.0;
      extrapolate_stb(&tangent, &h, &gamma, result);
    }

   free_vector(&h);
   free_vector(&tangent);
 }

/* Komplexe Ableitung berechnen */

void df(UBYTE *kernel, UBYTE *var, COMPLEX *z0, int diffticks, COMPLEX *result)
 {
   COMPLEX result1, result2;
   COMPLEX direc;
   COMPLEX temp1, temp2;
   COMPLEX error;

   direc.re = 0.0;		/* In Richtung der imaginaeren Achse */
   direc.im = 1.0;
   diff(kernel, var, z0, &direc, FALSE, diffticks, &result1);
   direc.re = 1.0;		/* In Richtung der reellen Achse */
   direc.im = 0.0;
   diff(kernel, var, z0, &direc, FALSE, diffticks, &result2);

   result->re = (result1.re + result2.re) / 2.0; 	/* Mittelwert bilden */
   result->im = (result1.im + result2.im) / 2.0;

   sub(result, &result1, &temp1);	/* Und Abweichung bestimmen */
   sub(result, &result2, &temp2);
   error.re = sqrt(complex_abs_sq(&temp1) + complex_abs_sq(&temp2));
   error.im = 0.0;

   assign_variable((UBYTE *) "error_df", &error);
 }

/* Eine Funktion minimieren */


void minmaximize(UBYTE *kernel, UBYTE *var, COMPLEX *z0, COMPLEX *result, int flag)
 {
   register int i;
   COMPLEX place1, place2, value1, value2;
   COMPLEX direc;

   direc.re = 1.0;
   direc.im = 0.0;
   place1.re = z0->re;
   place1.im = z0->im;
   place2.re = z0->re+direc.re;
   place2.im = z0->im+direc.im;

sloop:
 if(minoutput) {
   printf("Current position: ");
   show_complex(&place1);
   printf("Direction: ");
   show_complex(&direc);
     }
   assign_variable(var, &place1);
   evaluate(kernel, &value1);
   assign_variable(var, &place2);
   evaluate(kernel, &value2);
   if(complex_abs(&direc) < PRECISION)
    {
      result->re = place1.re;
      result->im = place1.im;
      return;
    }
   if(flag*value1.re > flag*value2.re)
    {
      place1.re = place2.re;
      place1.im = place2.im;
      place2.re = place1.re+direc.re;
      place2.im = place1.im+direc.im;
    }
   else
    {
      direc.re = -direc.re;
      direc.im = -direc.im;
      place2.re = place1.re+direc.re;
      place2.im = place1.im+direc.im;
   if(minoutput) {
      printf(" Current position: ");
      show_complex(&place1);
      printf(" Direction: ");
      show_complex(&direc);
          }
      assign_variable(var, &place1);
      evaluate(kernel, &value1);
      assign_variable(var, &place2);
      evaluate(kernel, &value2);
      if(flag*value1.re > flag*value2.re)
       {
         place1.re = place2.re;
         place1.im = place2.im;
         place2.re = place1.re+direc.re;
         place2.im = place1.im+direc.im;
       }
      else
       {
         direc.re = direc.re*0.5;
         direc.im = direc.im*0.5;
         place2.re = place1.re+direc.re;
         place2.im = place1.im+direc.im;
         goto sloop;
       }
    }
   

   while(1)
    {
   if(minoutput) {
     printf("  Current position: ");
     show_complex(&place1);
     printf("  Direction: ");
     show_complex(&direc);
       }
     assign_variable(var, &place1);
     evaluate(kernel, &value1);
     assign_variable(var, &place2);
     evaluate(kernel, &value2);
     if(complex_abs(&direc) < PRECISION)
      {
        result->re = place1.re;
        result->im = place1.im;
        return;
      }
     if(flag*value1.re > flag*value2.re)
      {
        place1.re = place2.re;
        place1.im = place2.im;
        place2.re = place1.re+direc.re;
        place2.im = place1.im+direc.im;
      }
     else
      {
        direc.re = -direc.re;
        direc.im = -direc.im;
        place2.re = place1.re+direc.re;
        place2.im = place1.im+direc.im;
      if(minoutput) {
        printf("   Current position: ");
        show_complex(&place1);
        printf("   Direction: ");
        show_complex(&direc);
         }
        assign_variable(var, &place1);
        evaluate(kernel, &value1);
        assign_variable(var, &place2);
        evaluate(kernel, &value2);
        if(flag*value1.re > flag*value2.re)
         {
           place1.re = place2.re;
           place1.im = place2.im;
           place2.re = place1.re+direc.re;
           place2.im = place1.im+direc.im;
         }
        else
         {
           direc.re = direc.re*0.5;
           direc.im = direc.im*0.5;
           place2.re = place1.re+direc.re;
           place2.im = place1.im+direc.im;
         }
      }
    }
 }
