/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "short_alloc.h"
#include "variables.h"
#include "text.h"
#include "matrix.h"
#include "arguments.h"
#include "vectors.h"
#include "apply_h.h"


/* *****************************
   Die Verwaltung von Matrizen */

/* Liste der Matrizen */

struct matrix *matrix_list=0;

/* Matrix-Elemente freigeben */

void free_matrix_elements(struct matrix_el *ptr)
 {
   struct matrix_el *hptr;

   for(; ptr; ptr=hptr)
    {
      hptr=ptr->next;
      short_free((BYTE *)ptr, sizeof(struct matrix_el));
    }
 }

/* Alle Matrizen freigeben */

void free_matrices()
 {
   register int i;
   struct matrix *ptr, *hptr;

   for(ptr=matrix_list; ptr; ptr=hptr)
    {
      hptr=ptr->next;
      if(ptr->text[0] == '*')		/* This is "the" internal matrix */
         exit_interal_h();		/* So, clean up */
      free(ptr->text);
      if(ptr->matrix_rows)
       {
/* We don't need to do this - we can do it all at once later ****
        for(i=0; i<ptr->rows; i++)
         {
          free_matrix_elements(ptr->matrix_rows[i]);
         }
*/
        free(ptr->matrix_rows);
       }
      free(ptr);
    }
/* Now free space occupied by all the matrix elements */
   short_freeall();

   matrix_list=0;
 }

/* Notification function for matrix allocation */

void void_notify_matrix(UBYTE *text)
 {
 }

void (*notify_ptr)(UBYTE *text) = void_notify_matrix;

/* Eine Matrix allokatieren */

void alloc_matrix(UBYTE *text, int rows, int columns)
 {
   struct matrix *ptr;
   register int i;

   if(text[0] != '*')
     if((rows < 1) || (columns < 1))
      {
        display_error("Illegal matrix arguments in definition");
        return;
      }

   (*notify_ptr)(text);

   for(ptr=matrix_list; ptr; ptr=ptr->next)	/* Does this matrix already exist ? */
     if(compare(text, ptr->text))
       {
        for(i=0; i<ptr->rows; i++)		/* Discard old contents of matrix */
          free_matrix_elements(ptr->matrix_rows[i]);
        free(ptr->matrix_rows);
						/* Allocate new structure */
        ptr->matrix_rows=(struct matrix_el **)malloc(rows * sizeof(struct matrix_el *));
        if(! ptr->matrix_rows)
          goto error;
        for(i=0; i<rows; i++)
            ptr->matrix_rows[i] = 0;
        ptr->rows=rows;
        ptr->columns = columns;
        return;
       }

   ptr = (struct matrix *)malloc(sizeof(struct matrix));
   if(! ptr)
     goto error;
   ptr->text = malloc(strlen(text) + 1);
   if(! ptr->text)
    {
      free(ptr);
      goto error;
    }

   for(i=0; ptr->text[i]=text[i]; i++)
           ;

   if(text[0] == '*')		/* Is it a matrix which we are going to treat internally ? */
    {
     ptr->rows = ptr->columns = init_interal_h();
     ptr->matrix_rows =0;
     ptr->next=matrix_list;
     matrix_list=ptr;
     return;
    }
     
   ptr->matrix_rows=(struct matrix_el **)malloc(rows * sizeof(struct matrix_el *));
   if(! ptr->matrix_rows)
    {
      free(ptr->text);
      free(ptr);
      goto error;
    }

   for(i=0; i<rows; i++)
        ptr->matrix_rows[i] = 0;
   ptr->rows=rows;
   ptr->columns = columns;
   ptr->next=matrix_list;
   matrix_list=ptr;
   return;
error:
   display_error("Out of memory while allocating new matrix");
 }

/* Notification function for matrix element overwriting */
 
void void_notify_matrix_overwrite()
 {
 }
 
void (*notify_ptr_overwrite)() = void_notify_matrix_overwrite;

/* Ein Matrix-Element einfuegen */

void insert_matrix_el(struct matrix_el **list, int column, COMPLEX *value)
 {
   struct matrix_el *ptr, *last_ptr;

   last_ptr = 0;
   for(ptr = *list; ptr; ptr=ptr->next)
    {
      if(ptr->column == column)
       {			/* Overwrite matrix element */
         (*notify_ptr_overwrite)();
         ptr->value.re = value->re;
         ptr->value.im = value->im;
         return;
       }
      if(ptr->column > column)
        goto alloc_it;
      last_ptr=ptr;
    }

 alloc_it:
   ptr=(struct matrix_el *)short_malloc(sizeof(struct matrix_el));
   if(! ptr)
    {
      display_error("Out of memory while assigning matrix element");
      return;
    }
   ptr->column = column;
   ptr->value.re = value->re;
   ptr->value.im = value->im;

   if(! last_ptr)
    {
      ptr->next = *list;
      *list = ptr;
    }
   else
    {
      ptr->next = last_ptr->next;
      last_ptr->next=ptr;
    }
 }

/* Matrix-Element zuweisen */

void assign_matrix_el(UBYTE *str, int row, int column, COMPLEX *value)
 {
   struct matrix *ptr;

   for(ptr=matrix_list; ptr; ptr=ptr->next)
    if( compare(str, ptr->text))
     {
      if((row < 1) || (column < 1) || (row > ptr->rows) || (column > ptr->columns))
       {
         display_error("Range exceeded during assignment of matrix element");
         return;
       }
      insert_matrix_el(&(ptr->matrix_rows[row-1]), column-1, value);
      return;
     }
   display_error_text("Unknown matrix '%s'", str);
 }

/* Eine Matrix nach Namen suchen */

struct matrix *identify_matrix(UBYTE *text)
 {
  struct matrix *ptr;

  for(ptr=matrix_list; ptr; ptr=ptr->next)
   {
    if(compare(text, ptr->text))
      return(ptr);
    if((text[0] == '*') && (ptr->text[0] == '*'))
      return(ptr);			/* Identification of internal matrix */
   }

  display_error_text("No matrix '%s' defined", text);
  return(0);
 }

/* Eine Matrix anzeigen */

void show_matrix(struct matrix *matrix)
 {
   register int i;
   struct matrix_el *ptr;

   for(i=0; i<matrix->rows; i++)
      for(ptr=matrix->matrix_rows[i]; ptr; ptr=ptr->next)
       {
         printf("%s(%d,%d) = ", matrix->text, i+1, ptr->column+1);
         show_complex(&(ptr->value));
       }

   printf(" ... no further nonzero matrix elements\n");
 }

/* Die Adjungierte einer Matrix erzeugen */

void adjoint_matrix(struct matrix *src, struct matrix *dst)
 {
   register int i;
   struct matrix_el *ptr;
   COMPLEX te;

   if((src->rows != dst->columns) || (src->columns != dst->rows))
    {
      display_error("Incorrect matrix sizes for taking adjoint");
      return;
    }

   for(i=0; i<src->rows; i++)
      for(ptr=src->matrix_rows[i]; ptr; ptr=ptr->next)
       {
         complex_conjugate(&(ptr->value), &te);
         insert_matrix_el(&(dst->matrix_rows[ptr->column]), i, &te);
       }
 }

/* Matrix-Argumente holen */

int get_matrix_var(UBYTE *instr, UBYTE *str, int *rows, int *columns)
 {
   int ofs;
   register int i;

   i=ofs=0;
   while(instr[ofs])
    {
     if(instr[ofs] > ' ')
        str[i++]=instr[ofs];
     ofs++;
    }
   str[i]=0;

   for(i=0; str[i]; i++)
    {
      if(str[i] == '(')
       {
         str[i++] = 0;
         i=get_number(str, i, rows);
         if(str[i++] != ',')
           goto error;
         i=get_number(str, i, columns);
         if(str[i++] != ')')
           goto error;
         if((! str[i]) || (str[i] == ';'))
           return(TRUE);
         goto error;
       }
    }
error:
   display_error("Illegal arguments of matrix");
   return(FALSE);
 }

/* Ein Matrix-Element auslesen */

int get_matrix_el(struct matrix *matrix, int row, int column, COMPLEX *result)
 {
   struct matrix_el *ptr;

   result->re = result->im = 0.0;

   if((row < 1) || (column < 1) || (row > matrix->rows) || (column > matrix->columns))
    {
      display_error("Range exceeded while retrieving matrix element");
      return(FALSE);
    }

   for(ptr = matrix->matrix_rows[row-1]; ptr; ptr=ptr->next)
    {
      if(ptr->column == column-1)
       {
         result->re = ptr->value.re;
         result->im = ptr->value.im;
         return(TRUE);
       }
      if(ptr->column > column-1)
         return(TRUE);
    }
   return(TRUE);
 }

/* ********************************
   A structure to store some temporary results
   for matrix operations */

struct mp_row {int flag;
               COMPLEX value; };

/* Add two matrices */

/* auxiliary functions */

static int ptr_column_lt(struct matrix_el *ptr, int j)
 {
  if(! ptr)
   return(FALSE);
  return(ptr->column < j);
 }

static int ptr_column_eq(struct matrix_el *ptr, int j)
 {
  if(! ptr)
   return(FALSE);
  return(ptr->column == j);
 }

void madd(struct matrix *src1, struct matrix *src2, struct matrix *dst)
 {
   register int i, j;
   struct matrix_el *ptr, *ptr1;
   COMPLEX temp;

   if((src1->rows != dst->rows) || (src1->columns != dst->columns) ||
      (src2->rows != dst->rows) || (src2->columns != dst->columns))
    {
      display_error("Incorrect matrix sizes for matrix addition");
      return;
    }

   for(i=0; i<dst->rows; i++)
    {
     ptr1 = src2->matrix_rows[i];		/* The actual addition */
     for(ptr=src1->matrix_rows[i]; ptr; ptr=ptr->next)
      {
        j = ptr->column;
        for(; ptr_column_lt(ptr1, j); ptr1=ptr1->next)
           insert_matrix_el(&(dst->matrix_rows[i]), ptr1->column, &(ptr1->value));
        if(ptr_column_eq(ptr1, j))
         {
          add(&(ptr->value), &(ptr1->value), &temp);
          insert_matrix_el(&(dst->matrix_rows[i]), j, &temp);
          ptr1=ptr1->next;
         }
        else
          insert_matrix_el(&(dst->matrix_rows[i]), j, &(ptr->value));
      }
     for(; ptr1; ptr1=ptr1->next)
        insert_matrix_el(&(dst->matrix_rows[i]), ptr1->column, &(ptr1->value));
    }
 }

/* Multiply two matrices */

void mprod(struct matrix *src1, struct matrix *src2, struct matrix *dst)
 {
   register int i, j;
   struct matrix_el *ptr, *ptr1;
   COMPLEX te;
   struct mp_row *temp_row;

   if((src1->rows != dst->rows) || (src2->columns != dst->columns))
    {
      display_error("Incorrect matrix sizes for matrix multiplication");
      return;
    }

   temp_row = (struct mp_row *) malloc(sizeof(struct mp_row) * dst->columns);
   if(! temp_row)
    {
     display_error("Not enough memory for temporary row during matrix multiplication");
     return;
    }

   for(i=0; i<dst->rows; i++)
    {
     for(j=0; j<dst->columns; j++)
      {
       temp_row[j].flag = FALSE;
       temp_row[j].value.re = temp_row[j].value.im = 0;
      }
     for(ptr=src1->matrix_rows[i]; ptr; ptr=ptr->next)    /* The actual multiplication */
       for(ptr1=src2->matrix_rows[ptr->column]; ptr1; ptr1=ptr1->next)
         {
          multiply(&(ptr->value), &(ptr1->value), &te);
          j = ptr1->column;
          add(&te, &(temp_row[j].value), &(temp_row[j].value));
          temp_row[j].flag = TRUE;
         }
     for(j=0; j<dst->columns; j++)
      if(temp_row[j].flag)
         insert_matrix_el(&(dst->matrix_rows[i]), j, &(temp_row[j].value)); 
    }
   free(temp_row);
 }

/* Commute two matrices */

void commute(struct matrix *src1, struct matrix *src2, struct matrix *dst)
 {
   register int i, j;
   struct matrix_el *ptr, *ptr1;
   COMPLEX te;
   struct mp_row *temp_row;

   if((src1->rows != dst->rows) || (src2->rows != dst->rows) ||
      (src1->columns != dst->columns) || (src2->columns != dst->columns) ||
      (src1->rows != src1->columns))
    {
      display_error("Incorrect matrix sizes for matrix commutation");
      return;
    }

   temp_row = (struct mp_row *) malloc(sizeof(struct mp_row) * dst->columns);
   if(! temp_row)
    {
     display_error("Not enough memory for temporary row during matrix commutation");
     return;
    }

   for(i=0; i<dst->rows; i++)
    {
     for(j=0; j<dst->columns; j++)
      {
       temp_row[j].flag = FALSE;
       temp_row[j].value.re = temp_row[j].value.im = 0;
      }
     for(ptr=src1->matrix_rows[i]; ptr; ptr=ptr->next)	/* The actual commutation */
       for(ptr1=src2->matrix_rows[ptr->column]; ptr1; ptr1=ptr1->next)
         {
          multiply(&(ptr->value), &(ptr1->value), &te);
          j = ptr1->column;
          add(&te, &(temp_row[j].value), &(temp_row[j].value));
          temp_row[j].flag = TRUE;
         }
     for(ptr=src2->matrix_rows[i]; ptr; ptr=ptr->next)	/* reverse order */
       for(ptr1=src1->matrix_rows[ptr->column]; ptr1; ptr1=ptr1->next)
         {
          multiply(&(ptr->value), &(ptr1->value), &te);
          j = ptr1->column;
          sub(&(temp_row[j].value), &te, &(temp_row[j].value));
          temp_row[j].flag = TRUE;
         }
     for(j=0; j<dst->columns; j++)
      if(temp_row[j].flag)
       if(temp_row[j].value.re || temp_row[j].value.im)
         insert_matrix_el(&(dst->matrix_rows[i]), j, &(temp_row[j].value));
    }
   free(temp_row);
 }
