/*  *** March 24-25, 2005:
    - parallelization of gr_sch1() routine
*/

/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "variables.h"
#include "text.h"
#include "matrix.h"
#include "vectors.h"
#include "arguments.h"
#include "set.h"

#include "apply_h.h"

/* Flag, ob im Skalarprodukt komplex konjugiert werden soll */

int sprod_complex;

/* *****************************
   Die Verwaltung von Vektoren */

/* Liste aller Vektoren */

static struct vector *vector_list = 0;

/* Alle externen Vektoren freigeben */

void free_vectors()
 {
   struct vector *ptr, *hptr;

   for(ptr=vector_list; ptr; ptr=hptr)
    {
      hptr=ptr->next;
      free(ptr->text);
      free(ptr->elements);
      free(ptr);
    }
   vector_list = 0;
 }

/* Inhalt eines Vektors freigeben */

void free_vector(struct vector *vec)
 {
   if(vec->elements)
     free(vec->elements);
   vec->elements = 0;
   vec->el_cnt=0;
 }

/* Free several internal vectors in a piece */

void free_Nvectors(struct vector *vec, int num)
 {
   register int i;

   if(vec[0].elements)
     free(vec[0].elements);

   for(i=0; i<num; i++)
    {
     vec[i].elements = 0;
     vec[i].el_cnt=0;
    }
 }


static char internal_text[] = "Internal vector";

/* Einen internen Vektor allokatieren */

void alloc_int_vector(struct vector *vec, int cnt)
 {
   register int i;

   vec->elements = 0;
   vec->el_cnt=0;
   if(cnt < 1)
    {
      display_error("Illegal vector arguments in definition");
      return;
    }

   vec->elements = (COMPLEX *) malign16(cnt * sizeof(COMPLEX));
   if(! vec->elements)
    {
      display_error("Out of memory while allocating vector");
      return;
    }
   for(i=0; i<cnt; i++)		/* Initialisierung auf 0 */
     vec->elements[i].re = vec->elements[i].im = 0.0;
   vec->el_cnt = cnt;

   vec->text = (UBYTE *) internal_text;
 }

/* Allocate several internal vectors in a block
   ATTENTION: Be very careful if you use this function:
              * Don't use functions such as move_int_vector()
                which spoil the blocking
              * Use free_Nvectors() to return allocated
                memory */ 

void alloc_int_vectors(struct vector *vec, int num, int cnt)
 {
   register int i, j;

   for(j=0; j<num; j++)
    {
     vec[j].elements = 0;
     vec[j].el_cnt=0;
    }
   if(cnt < 1)
    {
      display_error("Illegal vector arguments in definition");
      return;
    }

   vec[0].elements = (COMPLEX *) malign16(num * cnt * sizeof(COMPLEX));
   if(! vec[0].elements)
    {
      display_error("Out of memory while allocating vectors");
      return;
    }
#pragma omp parallel for private(i, j) if(cnt>500)
   for(j=0; j<num; j++)
    {
      if(j)
        vec[j].elements = vec[0].elements + j*cnt;
      for(i=0; i<cnt; i++)		/* Initialize to 0 */
        vec[j].elements[i].re = vec[j].elements[i].im = 0.0;
      vec[j].el_cnt = cnt;

      vec[j].text = (UBYTE *) internal_text;
    }
 }

/* Move an internal vector */

void move_int_vector(struct vector *src, struct vector *dst)
 {
  dst->elements = src->elements;
  dst->el_cnt = src->el_cnt;
  dst->text = src->text;

  src->elements=0;
  src->el_cnt=0;
 }

/* Einen Vektor allokatieren */

void alloc_vector(UBYTE *text, int cnt)
 {
   struct vector *ptr;
   register int i;

   if(cnt < 1)
    {
      display_error("Illegal vector arguments in definition");
      return;
    }

   for(ptr=vector_list; ptr; ptr=ptr->next)		/* Gibt es diesen Vektor schon ? */
     if(compare(text, ptr->text))
       {
         free(ptr->elements);
         ptr->el_cnt=0;
         ptr->elements = (COMPLEX *) malign16(cnt * sizeof(COMPLEX));
         if(! ptr->elements)
            goto error;
         for(i=0; i<cnt; i++)		/* Initialisierung auf 0 */
           ptr->elements[i].re = ptr->elements[i].im = 0.0;
         ptr->el_cnt = cnt;
         return;
       }

   ptr = (struct vector *)malloc(sizeof(struct vector));
   if(! ptr)
      goto error;

   ptr->elements = 0;
   ptr->el_cnt=0;

   ptr->text = malloc(strlen(text) + 1);		/* Text allokatieren */
   if(! ptr->text)
    {
      free(ptr);
      goto error;
    }
   ptr->elements = (COMPLEX *) malign16(cnt * sizeof(COMPLEX));
   if(! ptr->elements)
    {
      free(ptr->text);
      free(ptr);
      goto error;
    }
   for(i=0; i<cnt; i++)		/* Initialisierung auf 0 */
     ptr->elements[i].re = ptr->elements[i].im = 0.0;
   ptr->el_cnt = cnt;

   ptr->next = vector_list;		/* In Liste einfuegen */
   vector_list = ptr;

   for(i=0; ptr->text[i] = text[i]; i++)	/* Text kopieren */
        ;
   return;
error:
   display_error("Out of memory while allocating vector");
 }

/* Vektor-Komponente zuweisen */

void assign_vector_el(UBYTE *str, int row, COMPLEX *value)
 {
   struct vector *ptr;

   for(ptr=vector_list; ptr; ptr=ptr->next)
    if( compare(str, ptr->text))
     {
      if((row < 1) || (row > ptr->el_cnt))
       {
         display_error("Range exceeded during assignment of vector component");
         return;
       }
      ptr->elements[row-1].re = value->re;
      ptr->elements[row-1].im = value->im;
      return;
     }
   display_error_text("Unknown vector '%s'", str);
 }

/* Vektor-Argumente holen */

int get_vector_var(UBYTE *instr, UBYTE *str, int *rows)
 {
   int ofs;
   register int i;

   i=ofs=0;
   while(instr[ofs])
    {
     if(instr[ofs] > ' ')
        str[i++]=instr[ofs];
     ofs++;
    }
   str[i]=0;

   for(i=0; str[i]; i++)
    {
      if(str[i] == '(')
       {
         str[i++] = 0;
         i=get_number(str, i, rows);
         if(str[i++] != ')')
             goto error;
         if((! str[i]) || (str[i] == ';'))
             return(TRUE);
         goto error;
       }
    }
error:
   display_error("Illegal arguments of vector");
   return(FALSE);
 }

/* Einen Vektor nach Namen suchen */

struct vector *identify_vector(UBYTE *text, int error_flag)
 {
  struct vector *ptr;

  for(ptr=vector_list; ptr; ptr=ptr->next)
   if(compare(text, ptr->text))
     return(ptr);

  if(error_flag)
    display_error_text("No vector '%s' defined", text);
  return(0);
 }

/* Das Bild einer Operation finden bzw. initialisieren */

struct vector *init_dest_vector(UBYTE *name, int len)
 {
  struct vector *ptr;

  ptr = identify_vector(name, FALSE);
  if(ptr)
   {
    if(ptr->el_cnt != len)
     {
      display_error_text("Vector '%s' has wrong dimension", name);
      return(0);
     }
    return(ptr);
   }
  alloc_vector(name, len);
  ptr = identify_vector(name, FALSE);
  return(ptr);
 }

/* Einen Vektor ausgeben */

void show_vector(struct vector *vec)
 {
   register int i;

   for(i=0; i<vec->el_cnt; i++)
    {
      printf("%s(%d) = ", vec->text, i+1);
      show_complex(&(vec->elements[i]));
    }
 }

/* Einen Vektor kopieren */

void copy_vector(struct vector *source, struct vector *dest)
 {
   register int i;

   if(source->el_cnt != dest->el_cnt)
    {
       display_error("Cannot copy two vectors with unequal length");
       return;
    }
   for(i=0; i<dest->el_cnt; i++)
    {
       dest->elements[i].re=source->elements[i].re;
       dest->elements[i].im=source->elements[i].im;
    }
 }

/* Swap contents of two vectors */

void swap_vectors(struct vector *a, struct vector *b)
 {
   register int i;
   REAL re, im;

   if(a->el_cnt != b->el_cnt)
    {
       display_error("Cannot swap two vectors with unequal length");
       return;
    }
   for(i=0; i<b->el_cnt; i++)
    {
      re = a->elements[i].re;
      im = a->elements[i].im;
      a->elements[i].re=b->elements[i].re;
      a->elements[i].im=b->elements[i].im;
      b->elements[i].re = re;
      b->elements[i].im = im;
    }
 }


/* Skalarprodukt zweier Vektoren berechnen */

#ifdef SSE		/* Keep this for future SSE use */
void sprod(struct vector *x, struct vector *y, COMPLEX *result)
 {
   register int i;
   COMPLEX temp, temp1;

   result->re=result->im = 0;
   if(x->el_cnt != y->el_cnt)
    {
       display_error("Scalar product of two vectors with unequal length not defined");
       return;
    }
   if(sprod_complex)
    {
     for(i=0; i<x->el_cnt; i++)
      {
        temp1.re = x->elements[i].re;
        temp1.im = -x->elements[i].im;
        multiply(&temp1, &(y->elements[i]), &temp);
        add(&temp, result, result);
      }
    }
   else
    {
     for(i=0; i<x->el_cnt; i++)
      {
        multiply(&(x->elements[i]), &(y->elements[i]), &temp);
        add(&temp, result, result);
      }
    }
 }

#else
/* But otherwise use a version with better register usage */

void sprod(struct vector *x, struct vector *y, COMPLEX *result)
 {
   register int i, xcnt;
   register REAL resr, resi;
   register COMPLEX *xp, *yp;

   result->re=result->im = 0;
   xcnt = x->el_cnt;
   if(xcnt != y->el_cnt)
    {
       display_error("Scalar product of two vectors with unequal length not defined");
       return;
    }
   resr = resi = 0;
   xp = x->elements;
   yp = y->elements;
   if(sprod_complex)
    {
     for(i=0; i<xcnt; i++)
      {
        /* Complex conjugation implied */
        multiply_c_add_re(xp, yp, resr);
        multiply_c_add_im(xp, yp, resi);
        xp++;
        yp++;
      }
    }
   else
    {
     for(i=0; i<xcnt; i++)
      {
        multiply_add_re(xp, yp, resr);
        multiply_add_im(xp, yp, resi);
        xp++;
        yp++;
      }
    }

   result->re = resr;
   result->im = resi;
 }
#endif


/* Vielfaches eines Vektors von einem anderen abziehen */

#ifdef SSE		/* Keep this for future SSE use */
void sub_vector(COMPLEX *factor, struct vector *source, struct vector *dest)
 {
   register int i;
   COMPLEX temp;
   register REAL rfac;

   if(source->el_cnt != dest->el_cnt)
    {
       display_error("Sum of two vectors with unequal length not defined");
       return;
    }
   if(factor->im)
    {
     for(i=0; i<source->el_cnt; i++)
      {
        multiply(factor, &(source->elements[i]), &temp);
        sub(&(dest->elements[i]) ,&temp, &(dest->elements[i]));
      }
    }
   else
    {
     rfac = factor->re;
     for(i=0; i<source->el_cnt; i++)
      {
       dest->elements[i].re = dest->elements[i].re - rfac * source->elements[i].re;
       dest->elements[i].im = dest->elements[i].im - rfac * source->elements[i].im;
      }
    }
 }

#else
/* But otherwise use a version with better register usage */

void sub_vector(COMPLEX *factor, struct vector *source, struct vector *dest)
 {
   register int i, scnt;
   register REAL fr, fi;
   register COMPLEX *sp, *dp;

   scnt = source->el_cnt;
   if(scnt != dest->el_cnt)
    {
       display_error("Sum of two vectors with unequal length not defined");
       return;
    }
   fr = factor->re;
   fi = factor->im;
   sp = source->elements;
   dp = dest->elements;
   if(fi)
    {
     for(i=0; i<scnt; i++)
      {
       dp->re -= (fr * sp->re - fi * sp->im);
       dp->im -= (fi * sp->re + fr * sp->im);
       sp++;
       dp++;
      }
    }
   else
    {
     for(i=0; i<scnt; i++)
      {
       dp->re -= (fr * sp->re);
       dp->im -= (fr * sp->im);
       sp++;
       dp++;
      }
    }
 }
#endif

/* Subtract a vector without prefactor */

void sub_vector_nf(struct vector *source, struct vector *dest)
 {
   register int i, scnt;
    register COMPLEX *sp, *dp;

   scnt = source->el_cnt;
   if(scnt != dest->el_cnt)
    {
       display_error("Sum of two vectors with unequal length not defined");
       return;
    }
   sp = source->elements;
   dp = dest->elements;
   for(i=0; i<scnt; i++)
    {
     dp->re -= sp->re;
     dp->im -= sp->im;
     sp++;
     dp++;
    }
 }

/* Vielfaches eines Vektors zu einem anderen addiern */

#ifdef SSE		/* Keep this for future SSE use */
void add_vector(COMPLEX *factor, struct vector *source, struct vector *dest)
 {
   register int i;
   COMPLEX temp;

   if(source->el_cnt != dest->el_cnt)
    {
       display_error("Sum of two vectors with unequal length not defined");
       return;
    }
   for(i=0; i<source->el_cnt; i++)
    {
      multiply(factor, &(source->elements[i]), &temp);
      add(&(dest->elements[i]) ,&temp, &(dest->elements[i]));
    }
 }

#else
/* But otherwise use a version with better register usage */

void add_vector(COMPLEX *factor, struct vector *source, struct vector *dest)
 {
   register int i, scnt;
   register REAL fr, fi;
   register COMPLEX *sp, *dp;

   scnt = source->el_cnt;
   if(scnt != dest->el_cnt)
    {
       display_error("Sum of two vectors with unequal length not defined");
       return;
    }
   fr = factor->re;
   fi = factor->im;
   sp = source->elements;
   dp = dest->elements;
   for(i=0; i<scnt; i++)
    {
      dp->re += (fr * sp->re - fi * sp->im);
      dp->im += (fi * sp->re + fr * sp->im);
      sp++;
      dp++;
    }
 }
#endif


/* reelles vielfaches eines Vektors zu einem anderen addiern */

void add_vector_r(REAL factor, struct vector *source, struct vector *dest)
 {
   register int i;
   COMPLEX temp;

   if(source->el_cnt != dest->el_cnt)
    {
       display_error("Sum of two vectors with unequal length not defined");
       return;
    }
   for(i=0; i<source->el_cnt; i++)
    {
      temp.re = factor*source->elements[i].re;
      temp.im = factor*source->elements[i].im;
      add(&(dest->elements[i]) ,&temp, &(dest->elements[i]));
    }
 }

/* Zwei Vektoren zueinander addieren */

void add_vectors(struct vector *source1, struct vector *source2, struct vector *dest)
 {
   register int i;

   if(source1->el_cnt != source2->el_cnt)
    {
       display_error("Sum of two vectors with unequal length not defined");
       return;
    }
   if(source1->el_cnt != dest->el_cnt)
    {
       display_error("Sum of vectors must have same dimension as summands");
       return;
    }
   for(i=0; i<source1->el_cnt; i++)
      add(&(source1->elements[i]), &(source2->elements[i]), &(dest->elements[i]));
 }

/* Vektor mit einem Skalar multiplizieren */

void smpy_vector(COMPLEX *factor, struct vector *source, struct vector *dest)
 {
   register int i;
   COMPLEX temp;

   if(source->el_cnt != dest->el_cnt)
    {
       display_error("Source and destination must be the same when multiplying with a scalar");
       return;
    }
   for(i=0; i<source->el_cnt; i++)
    {
      multiply(factor, &(source->elements[i]), &temp);
      dest->elements[i].re = temp.re;
      dest->elements[i].im = temp.im;
    }
 }

/* Kreuz-Produkt dreier Vektoren */

void cross_product(struct vector *source1, struct vector *source2, struct vector *dest)
 {
   COMPLEX a,b;

   if((source1->el_cnt != 3) || (source2->el_cnt != 3) || (dest->el_cnt != 3))
    {
     display_error("The cross product is defined in three dimensions only");
     return;
    }

   multiply(source1->elements+1, source2->elements+2, &a);
   multiply(source1->elements+2, source2->elements+1, &b);
   sub(&a, &b, dest->elements);
   multiply(source1->elements+2, source2->elements, &a);
   multiply(source1->elements, source2->elements+2, &b);
   sub(&a, &b, dest->elements+1);
   multiply(source1->elements, source2->elements+1, &a);
   multiply(source1->elements+1, source2->elements, &b);
   sub(&a, &b, dest->elements+2);
 }

/* Norm eines Vektors berechnen */

REAL norm(struct vector *x)
 {
   register int i;
   REAL sum;

   sum = 0.0;
   for(i=0; i<x->el_cnt; i++)
     sum += complex_abs_sq(&(x->elements[i]));
   return(sqrt(sum));
 }

/* Einen Vektor auf 1 normieren */

void normalize_vector(struct vector *x)
 {
   register int i;
   REAL normv;

   normv = norm(x);
/*   printf("Norm = %17.17g\n", normv);  */
   if(normv == 0.0)
    {
      display_error("Cannot normalize null vector");
      return;
    }
   for(i=0; i<x->el_cnt; i++)
    {
      x->elements[i].re /= normv;
      x->elements[i].im /= normv;
    }
 }

/* Apply named matrix to vector */
/* *** August 1, 2002: Scalar variable temp1
       introduced result of summation
   *** November 10, 2004: Temporary storage
       temp eliminated */

#ifdef SSE		/* Keep this for future SSE use */
void apply_named_matrix(char *name, struct matrix *matrix,
                        struct vector *source, struct vector *dest)
 {
   register int i;
   struct matrix_el *ptr;
   COMPLEX temp1;

   if(matrix->columns != source->el_cnt)
    {
      display_error("Preimage of matrix has wrong dimension");
      return;
    }
   if(matrix->rows != dest->el_cnt)
    {
      display_error("Image of matrix has wrong dimension");
      return;
    }

   if(name[0] == '*')			/* This is an internal matrix */
    {
     apply_internal_ham(source, dest, name+1);
					/* So, call the appropriate routine */
     return;
    }

   for(i=0; i<matrix->rows; i++)
    {
      temp1.re = temp1.im = 0;
      for(ptr=matrix->matrix_rows[i]; ptr; ptr=ptr->next)
       {
         multiply_add(&(ptr->value), &(source->elements[ptr->column]), &temp1);
       }
      dest->elements[i].re = temp1.re;
      dest->elements[i].im = temp1.im;
    }
 }

#else
/* But otherwise use a version with better register usage */

void apply_named_matrix(char *name, struct matrix *matrix,
                        struct vector *source, struct vector *dest)
 {
   register int i;
   register struct matrix_el *ptr;
   register REAL tempr, tempi;

   if(matrix->columns != source->el_cnt)
    {
      display_error("Preimage of matrix has wrong dimension");
      return;
    }
   if(matrix->rows != dest->el_cnt)
    {
      display_error("Image of matrix has wrong dimension");
      return;
    }

   if(name[0] == '*')			/* This is an internal matrix */
    {
     apply_internal_ham(source, dest, name+1);
					/* So, call the appropriate routine */
     return;
    }

   for(i=0; i<matrix->rows; i++)
    {
      tempr = tempi = 0;
      for(ptr=matrix->matrix_rows[i]; ptr; ptr=ptr->next)
       {
         multiply_add_re(&(ptr->value), &(source->elements[ptr->column]), tempr);
         multiply_add_im(&(ptr->value), &(source->elements[ptr->column]), tempi);
       }
      dest->elements[i].re = tempr;
      dest->elements[i].im = tempi;
    }
 }

#endif

/* Apply matrix to vector */

void apply_matrix(struct matrix *matrix, struct vector *source, struct vector *dest)
 {
  apply_named_matrix(matrix->text, matrix, source, dest);
 }

/* ***************************************
   Eigenwert aus Rayleigh-Quotient schaetzen */

REAL estimate_eigenvalue(struct vector *vec1, struct vector *vec2, COMPLEX *result)
 {
   register int i;
   REAL diff;
   REAL normv;
   COMPLEX temp, temp1;

   normv = norm_nopar(vec1);
   temp1.re = normv * normv;
   temp1.im = 0.0;
   sprod_nopar(vec1, vec2, &temp);
   divide(&temp, &temp1, result);

   if(vec1->el_cnt != vec2->el_cnt)
      return(0.0);

   diff=0.0;
   for(i=0; i<vec1->el_cnt; i++)
    {
      divide(&(vec2->elements[i]), result, &temp);
      sub(&(vec1->elements[i]), &temp, &temp1);
/*      divide(&(vec2->elements[i]), &(vec1->elements[i]), &temp);
      sub(result, &temp, &temp1);  */
      diff += complex_abs_sq(&temp1);
    }
   return(diff);
 }


/* Vektor-Iteration */

void vector_iterate(struct matrix *m, struct vector *v, COMPLEX *shift, COMPLEX *prec, COMPLEX *result)
 {
   struct vector tempvec;
   REAL precision;
   REAL diff;
   COMPLEX value;
   int pass;
   register int i;

   if(m->rows != v->el_cnt)
    {
      display_error("Vector for vector iteration has wrong dimension");
      return;
    }

   if(m->rows != m->columns)
    {
      display_error("Eigenvalue problem only well-defined for square matrices");
      return;
    }

   precision = complex_abs(prec);

   alloc_int_vector(&tempvec, m->columns);

   if(norm(v) < 1e-16)
    {
      for(i=0; i<v->el_cnt; i++)
         v->elements[i].re = 1.0;
    }

   diff = 1e10;
   pass=0;
   while(diff > precision)
    {
     for(i=0; i<4; i++)
        {
          apply_matrix(m, v, &tempvec);
          sub_vector(shift, v, &tempvec);
          apply_matrix(m, &tempvec, v);
          sub_vector(shift, &tempvec, v);
          normalize_vector(v);
        }
     apply_matrix(m, v, &tempvec);
     sub_vector(shift, v, &tempvec);
     diff = sqrt(estimate_eigenvalue(v, &tempvec, &value));
     apply_matrix(m, &tempvec, v);
     sub_vector(shift, &tempvec, v);
     normalize_vector(v);

     add(&value, shift, result);
     if(OUTPUT_FLAG)
         printf("%d. Estimate for Eigenvalue: (%17.17g) ", pass*10+1, diff);
     show_complex(result);

     pass++;
    }
   free_vector(&tempvec);
 }

/* Orthonormierung von Vektoren nach Gram-Schmidt */

void gr_sch(struct vector *vec_list, int cnt, int start_with)
 {
   register int i, j;
   COMPLEX temp;

   for(j=0; j<cnt; j++)
    {
      if(j >= start_with)
        normalize_vector(vec_list+j);		/* Normalize to 1 */
#pragma omp parallel for private(i,temp)
      for(i=(j>=start_with)?(j+1):start_with; i<cnt; i++)
       {
         sprod_nopar(vec_list+j, vec_list+i, &temp);	/* Projektion berechnen */
         sub_vector_nopar(&temp, vec_list+j, vec_list+i);	/* Und abziehen */
       }
    }
 }

/* Orthonormalize one vector */

/* This variant works, but is less efficient than the next one below
   (cache use ?) */
#ifdef XXX
void gr_sch1(struct vector *vec_list, int cnt)
 {
   register int i, j;
   int dim;
   struct vector sp;
   register REAL sum_re, sum_im;

   if(cnt > 0)
     normalize_vector(vec_list+cnt-1);		/* Normalize to 1 */

   if(cnt > 1)
    {
     alloc_int_vector(&sp, cnt-1);    
#pragma omp parallel for private(j)
     for(j=0; j<cnt-1; j++)
       sprod_nopar(vec_list+j, vec_list+cnt-1, sp.elements+j);
     dim = vec_list[cnt-1].el_cnt;
#pragma omp parallel for private(i,j,sum_re,sum_im)
     for(i=0; i<dim; i++)
      {
       sum_re = sum_im = 0;
       for(j=0; j<cnt-1; j++)
        {
         multiply_add_re(sp.elements+j, vec_list[j].elements+i, sum_re);
         multiply_add_im(sp.elements+j, vec_list[j].elements+i, sum_im);
        }
       vec_list[cnt-1].elements[i].re -= sum_re;
       vec_list[cnt-1].elements[i].im -= sum_im;
      }
     free_vector(&sp);
    }
 }
#endif

#ifdef _OPENMP
extern int omp_numproc;
#define MAXPROC 36				/* maximal number of processors we can handle */

void gr_sch1(struct vector *vec_list, int cnt)
 {
   int np;
   register int j, proc;
   COMPLEX temp;
   struct vector tvec[MAXPROC];

   if(cnt > 0)
     normalize_vector(vec_list+cnt-1);		/* Normalize to 1 */

   if(cnt < 2)					/* only one vector ? */
     return;					/* -> we are done */

   np = omp_numproc;				/* number of threads */
   if(np > cnt-1)
     np = cnt-1;				/* not bigger than number of vectors */
   if(np >= MAXPROC)
     np = MAXPROC-1;				/* not bigger than maximal number of processors */

   if(cnt <= np+np)				/* for a few vectors, the parallelization overhead doesn't pay off */
    {
     for(j=0; j<cnt-1; j++)
      {
       sprod(vec_list+j, vec_list+cnt-1, &temp);		/* compute projection */
       sub_vector(&temp, vec_list+j, vec_list+cnt-1);		/* and subtract it */
      }
     return;
    }

   /* Allocate a temporary vector fo reach processor */
   for(proc=0; proc<np; proc++)
     alloc_int_vector(tvec+proc, vec_list->el_cnt);

#pragma omp parallel for private(proc,j,temp)
   for(proc=0; proc<np; proc++)
    for(j=proc; j<cnt-1; j += np)
     {
      sprod_nopar(vec_list+j, vec_list+cnt-1, &temp);		/* compute projection */
      add_vector_nopar(&temp, vec_list+j, tvec+proc);		/* and accumulate it */
     }

   /* Now collect results */
   for(proc=0; proc<np; proc++)
    {
     sub_vector_nf(tvec+proc, vec_list+cnt-1);		/* subtract it */
     free_vector(tvec+proc);				/* and free temporary storage */
    }
  }

#else

void gr_sch1(struct vector *vec_list, int cnt)
 {
   register int j;
   COMPLEX temp;

   if(cnt > 0)
     normalize_vector(vec_list+cnt-1);		/* Normalize to 1 */

   for(j=0; j<cnt-1; j++)
    {
     sprod(vec_list+j, vec_list+cnt-1, &temp);		/* compute projection */
     sub_vector(&temp, vec_list+j, vec_list+cnt-1);	/* and subtract it */
    }
 }

#endif


/* Solve linear system of equations using conjugate gradient method */
/* compare p. 215 - 217 of
   A. Jennings, "Matrix Computation for Engineers and Scientists,
   John Wiley & Sons, 1977 */

void cg_solve(struct matrix *m, struct vector *x, struct vector *b, COMPLEX *prec)
 {
  REAL precision;
  struct vector u, p, r;
  COMPLEX alpha, beta;
  COMPLEX norm_r, sp_t;
  REAL residue;
  int total_pass, cg_pass;

  if(m->rows != m->columns)
   {
     display_error("Conjugate gradient method works only for square matrices");
     return;
   }

  if(m->rows != b->el_cnt)
   {
     display_error("Vector on r.h.s. of equation has wrong dimension");
     return;
   }

  if(m->rows != x->el_cnt)
   {
     display_error("Vector on l.h.s. of equation has wrong dimension");
     return;
   }

  precision = complex_abs(prec);

  alloc_int_vector(&u, m->columns);
  alloc_int_vector(&p, m->columns);
  alloc_int_vector(&r, m->columns);

  total_pass = 0;
  cg_pass = 0;
  residue = 1e10;

  while(residue > precision*precision)
   {
    if(cg_pass >= CG_ITER)
     {
      if(OUTPUT_FLAG)
        printf("Restarting conjugate gradient method\n");
      cg_pass = 0;
     }
    if(! cg_pass)
     {					/* Initialization of cg iteration */
      copy_vector(b, &r);		/* eq. (6.120) of Jennings */
      apply_matrix(m, x, &u);
      add_vector_r(-1, &u, &r);
      copy_vector(&r, &p);
     }
    apply_matrix(m, &p, &u);            /* Actual cg iteration */
    sprod(&r, &r, &norm_r);             /* eq. (6.119) of Jennings */
    sprod(&p, &u, &sp_t);
    divide(&norm_r, &sp_t, &alpha);

    if((alpha.im > 1e-16) || (alpha.im <-1e-16))
      display_error("alpha is not real - cg_solve only works for hermitean matrices");

    add_vector_r(alpha.re, &p, x);
    add_vector_r(-alpha.re, &u, &r);
    sprod(&r, &r, &sp_t);
    divide(&sp_t, &norm_r, &beta); 
    smpy_vector(&beta, &p, &p);
    add_vectors(&p, &r, &p);

    residue = alpha.re * norm_r.re;     /* Improvement of residue */
                                        /* eq. (6.128) of Jennings */
    cg_pass++;
    total_pass++;

    if(OUTPUT_FLAG)
      printf("Improvement of residue square during %d. iteration (%d. cg-iteration) : %lg\n",
              total_pass, cg_pass, residue);
   }

  free_vector(&r);
  free_vector(&p);
  free_vector(&u);
 }

/* Solve linear system of equations using biconjugate gradient method */
/* compare Sec. 2.7 of "Numerical Recipes in C", W.H. Press, S.A.
   Teukolsky, W.T. Vetterling, B.P. Flannery, Cambridge University Press (1992) */

/* Name for temporary adjoint matrix */

static UBYTE BCG_AM_NAME[] = "__bcg_a_mat";

void bcg_solve(struct matrix *m, struct vector *x, struct vector *b, COMPLEX *prec)
 {
  REAL precision;
  struct vector u, p, r, ut, pt, rt;
  struct matrix *amat;
  COMPLEX alpha, beta;
  COMPLEX norm_r, sp_t;
  COMPLEX residue;
  int total_pass, cg_pass;

  if(m->rows != m->columns)
   {
     display_error("Biconjugate gradient method works only for square matrices");
     return;
   }

  if(m->rows != b->el_cnt)
   {
     display_error("Vector on r.h.s. of equation has wrong dimension");
     return;
   }

  if(m->rows != x->el_cnt)
   {
     display_error("Vector on l.h.s. of equation has wrong dimension");
     return;
   }

  alloc_matrix(BCG_AM_NAME, m->columns, m->rows);
  amat = identify_matrix(BCG_AM_NAME);
  if(amat)
   {
    adjoint_matrix(m, amat);

    precision = complex_abs(prec);

    alloc_int_vector(&u, m->columns);
    alloc_int_vector(&p, m->columns);
    alloc_int_vector(&r, m->columns);
    alloc_int_vector(&ut, m->columns);
    alloc_int_vector(&pt, m->columns);
    alloc_int_vector(&rt, m->columns);

    total_pass = 0;
    cg_pass = 0;
    residue.re = 1e10;
    residue.im = 0;

    while(complex_abs(&residue) > precision*precision)
     {
      if(cg_pass >= CG_ITER)
       {
        if(OUTPUT_FLAG)
          printf("Restarting biconjugate gradient method\n");
        cg_pass = 0;
       }
      if(! cg_pass)
       {				/* Initialization of cg iteration */
        copy_vector(b, &r);
        apply_matrix(m, x, &u);
        add_vector_r(-1, &u, &r);
        copy_vector(&r, &p);
        copy_vector(&r, &rt);
        copy_vector(&r, &pt);
       }
      apply_matrix(m, &p, &u);		/* Actual bcg iteration */
      apply_matrix(amat, &pt, &ut);
      sprod(&rt, &r, &norm_r);
      sprod(&pt, &u, &sp_t);
      divide(&norm_r, &sp_t, &alpha);

      multiply(&alpha, &norm_r, &residue);	/* Improvement of residue */

      add_vector(&alpha, &p, x);
      sub_vector(&alpha, &u, &r);
      alpha.im = -alpha.im;		/* alpha^* */
      sub_vector(&alpha, &ut, &rt);
      sprod(&rt, &r, &sp_t);
      divide(&sp_t, &norm_r, &beta); 
      smpy_vector(&beta, &p, &p);
      add_vectors(&p, &r, &p);
      beta.im = -beta.im;		/* beta^* */
      smpy_vector(&beta, &pt, &pt);
      add_vectors(&pt, &rt, &pt);

      cg_pass++;
      total_pass++;

      if(OUTPUT_FLAG)
        printf("Improvement of residue square during %d. iteration (%d. bcg-iteration) : %lg\n",
               total_pass, cg_pass, complex_abs(&residue));
     }

    free_vector(&rt);
    free_vector(&pt);
    free_vector(&ut);
    free_vector(&r);
    free_vector(&p);
    free_vector(&u);
    /* If we could, we should free the space occupied by the adjoint
       matrix ... */
   }
 }
