/* Definitions for Operations on vectors */

/* Structure of a vector */

struct vector { struct vector *next;
                UBYTE *text;
                int el_cnt;
                COMPLEX *elements;
              };

/* Flag if complex conjugation is to be used in scalar product */

extern int sprod_complex;

/* Functions for vectors */

extern void free_vectors();
extern void free_vector(struct vector *vec);
extern void free_Nvectors(struct vector *vec, int num);
extern void alloc_int_vector(struct vector *vec, int cnt);
extern void alloc_int_vectors(struct vector *vec, int num, int cnt);
extern void move_int_vector(struct vector *src, struct vector *dst);
extern void alloc_vector(UBYTE *text, int cnt);
extern void assign_vector_el(UBYTE *str, int row, COMPLEX *value);
extern int get_vector_var(UBYTE *instr, UBYTE *str, int *rows);
extern struct vector *identify_vector(UBYTE *text, int error_flag);
extern struct vector *init_dest_vector(UBYTE *name, int len);
extern void show_vector(struct vector *vec);
extern void copy_vector(struct vector *source, struct vector *dest);
extern void swap_vectors(struct vector *a, struct vector *b);
extern void sprod(struct vector *x, struct vector *y, COMPLEX *result);
extern void sub_vector(COMPLEX *factor, struct vector *source, struct vector *dest);
extern void sub_vector_nf(struct vector *source, struct vector *dest);
extern void add_vector(COMPLEX *factor, struct vector *source, struct vector *dest);
extern void add_vector_r(REAL factor, struct vector *source, struct vector *dest);
extern void add_vectors(struct vector *source1, struct vector *source2, struct vector *dest);
extern void smpy_vector(COMPLEX *factor, struct vector *source, struct vector *dest);
extern void cross_product(struct vector *source1, struct vector *source2, struct vector *dest);
extern REAL norm(struct vector *x);
extern void normalize_vector(struct vector *x);
extern void apply_matrix(struct matrix *matrix, struct vector *source, struct vector *dest);
extern void apply_named_matrix(char *name, struct matrix *matrix, struct vector *source, struct vector *dest);
extern void vector_iterate(struct matrix *m, struct vector *v, COMPLEX *shift, COMPLEX *prec, COMPLEX *result);
extern void gr_sch(struct vector *vec_list, int cnt, int start_with);
extern void gr_sch1(struct vector *vec_list, int cnt);
extern REAL estimate_eigenvalue(struct vector *vec1, struct vector *vec2, COMPLEX *result);
extern void cg_solve(struct matrix *m, struct vector *x, struct vector *b, COMPLEX *prec);
extern void bcg_solve(struct matrix *m, struct vector *x, struct vector *b, COMPLEX *prec);

#define sprod_nopar(a,b,c) sprod(a,b,c)
#define sub_vector_nopar(a,b,c) sub_vector(a,b,c)
#define add_vector_nopar(a,b,c) add_vector(a,b,c)
#define norm_nopar(a) norm(a)
#define normalize_vector_nopar(a) normalize_vector(a)
#define apply_matrix_nopar(a,b,c) apply_matrix(a,b,c)
